# SPDX-License-Identifier: MIT

import os
import gc
import shutil
import numpy as np
from mock import MagicMock

from ..utility import QtTest, click_button, click_checkbox

from qtpy import QtCore, QtWidgets
from qtpy.QtTest import QTest

from ...model.util.HelperModule import get_partial_value
from ...widgets.integration import IntegrationWidget
from ...controller.integration.ImageController import ImageController
from ...model.DioptasModel import DioptasModel

unittest_data_path = os.path.join(os.path.dirname(__file__), "../data")


def test_automatic_file_processing(
    integration_widget, dioptas_model, image_controller, qapp
):
    # get into a specific folder
    QtWidgets.QFileDialog.getOpenFileNames = MagicMock(
        return_value=[os.path.join(unittest_data_path, "image_001.tif")]
    )
    click_button(integration_widget.load_img_btn)
    assert str(integration_widget.img_filename_txt.text()) == "image_001.tif"
    assert dioptas_model.working_directories["image"] == unittest_data_path

    click_checkbox(integration_widget.autoprocess_cb)

    assert not dioptas_model.img_model._directory_watcher.signalsBlocked()
    assert integration_widget.autoprocess_cb.isChecked()
    assert dioptas_model.img_model.autoprocess

    shutil.copy2(
        os.path.join(unittest_data_path, "image_001.tif"),
        os.path.join(unittest_data_path, "image_003.tif"),
    )

    dioptas_model.img_model._directory_watcher.file_added.emit(
        os.path.join(unittest_data_path, "image_003.tif")
    )
    assert "image_003.tif" == str(integration_widget.img_filename_txt.text())
    qapp.processEvents()

    # clean up
    os.remove(os.path.join(unittest_data_path, "image_003.tif"))


def test_configuration_selected_changes_mask_mode(
    integration_widget, dioptas_model, image_controller
):
    dioptas_model.add_configuration()
    click_button(integration_widget.img_mask_btn)
    assert dioptas_model.use_mask

    dioptas_model.select_configuration(0)
    assert not dioptas_model.use_mask
    assert not integration_widget.img_mask_btn.isChecked()


def test_configuration_selected_changes_mask_transparency(
    integration_widget, dioptas_model, image_controller
):
    click_button(integration_widget.img_mask_btn)
    dioptas_model.add_configuration()
    click_button(integration_widget.img_mask_btn)
    click_checkbox(integration_widget.mask_transparent_cb)
    assert dioptas_model.transparent_mask

    dioptas_model.select_configuration(0)
    assert not dioptas_model.transparent_mask
    assert not integration_widget.mask_transparent_cb.isChecked()


def test_configuration_selected_changed_autoprocessing_of_images(
    integration_widget, dioptas_model, image_controller
):
    click_checkbox(integration_widget.autoprocess_cb)
    dioptas_model.add_configuration()

    assert not dioptas_model.img_model.autoprocess
    assert not integration_widget.autoprocess_cb.isChecked()

    dioptas_model.select_configuration(0)
    assert dioptas_model.img_model.autoprocess
    assert integration_widget.autoprocess_cb.isChecked()


def test_configuration_selected_changes_calibration_name(
    integration_widget, dioptas_model, image_controller
):
    dioptas_model.calibration_model.calibration_name = "calib1"
    dioptas_model.add_configuration()
    dioptas_model.calibration_model.calibration_name = "calib2"

    dioptas_model.select_configuration(0)
    assert str(integration_widget.calibration_lbl.text()) == "calib1"

    dioptas_model.select_configuration(1)
    assert str(integration_widget.calibration_lbl.text()) == "calib2"


def test_configuration_selected_updates_mask_plot(
    integration_widget, dioptas_model, image_controller
):
    dioptas_model.mask_model.add_mask(os.path.join(unittest_data_path, "test.mask"))
    click_button(integration_widget.img_mask_btn)
    first_mask = dioptas_model.mask_model.get_img()
    dioptas_model.add_configuration()
    assert not integration_widget.img_mask_btn.isChecked()

    dioptas_model.mask_model.mask_below_threshold(dioptas_model.img_data, 1)
    second_mask = dioptas_model.mask_model.get_img()
    click_button(integration_widget.img_mask_btn)

    dioptas_model.select_configuration(0)
    assert np.sum(integration_widget.img_widget.mask_img_item.image - first_mask) == 0
    dioptas_model.select_configuration(1)
    assert np.sum(integration_widget.img_widget.mask_img_item.image - second_mask) == 0


def test_configuration_selected_updates_mask_transparency(
    integration_widget, dioptas_model, image_controller
):
    dioptas_model.mask_model.add_mask(os.path.join(unittest_data_path, "test.mask"))
    click_button(integration_widget.img_mask_btn)
    dioptas_model.add_configuration()

    assert not integration_widget.img_mask_btn.isChecked()
    assert not integration_widget.mask_transparent_cb.isVisible()


def test_configuration_selected_updates_roi_mode(
    integration_widget, dioptas_model, image_controller
):
    click_button(integration_widget.img_roi_btn)
    assert integration_widget.img_roi_btn.isChecked()
    assert (
        integration_widget.img_widget.roi
        in integration_widget.img_widget.img_view_box.addedItems
    )

    dioptas_model.add_configuration()
    assert not integration_widget.img_roi_btn.isChecked()
    assert (
        not integration_widget.img_widget.roi
        in integration_widget.img_widget.img_view_box.addedItems
    )


def test_mask_button_checking_in_image_mode(integration_widget, image_controller):
    click_button(integration_widget.img_mask_btn)
    assert integration_widget.img_mask_btn.isChecked()
    click_button(integration_widget.img_mask_btn)
    assert not integration_widget.img_mask_btn.isChecked()


def test_mask_button_checking_in_cake_mode(
    integration_widget, dioptas_model, image_controller
):
    load_pilatus1M_image_and_calibration(dioptas_model)
    click_button(integration_widget.integration_image_widget.mode_btn)

    click_button(integration_widget.img_mask_btn)
    assert dioptas_model.use_mask
    assert integration_widget.img_mask_btn.isChecked()
    click_button(integration_widget.img_mask_btn)
    assert not dioptas_model.use_mask
    assert not integration_widget.img_mask_btn.isChecked()


def test_adding_images(integration_widget, image_controller):
    QtWidgets.QFileDialog.getOpenFileNames = MagicMock(
        return_value=[os.path.join(unittest_data_path, "image_001.tif")]
    )
    click_button(integration_widget.load_img_btn)
    data1 = np.copy(integration_widget.img_widget.img_data).astype(np.uint32)
    click_checkbox(integration_widget.img_batch_mode_add_rb)
    QtWidgets.QFileDialog.getOpenFileNames = MagicMock(
        return_value=[
            os.path.join(unittest_data_path, "image_001.tif"),
            os.path.join(unittest_data_path, "image_001.tif"),
        ]
    )
    click_button(integration_widget.load_img_btn)
    assert np.array_equal(2 * data1, integration_widget.img_widget.img_data)


def test_average_images(integration_widget, image_controller):
    QtWidgets.QFileDialog.getOpenFileNames = MagicMock(
        return_value=[os.path.join(unittest_data_path, "image_001.tif")]
    )
    click_button(integration_widget.load_img_btn)
    data1 = np.copy(integration_widget.img_widget.img_data).astype(np.uint32)

    QtWidgets.QFileDialog.getOpenFileNames = MagicMock(
        return_value=[os.path.join(unittest_data_path, "image_002.tif")]
    )
    click_button(integration_widget.load_img_btn)
    data2 = np.copy(integration_widget.img_widget.img_data).astype(np.uint32)

    click_checkbox(integration_widget.img_batch_mode_average_rb)
    QtWidgets.QFileDialog.getOpenFileNames = MagicMock(
        return_value=[
            os.path.join(unittest_data_path, "image_001.tif"),
            os.path.join(unittest_data_path, "image_002.tif"),
        ]
    )
    click_button(integration_widget.load_img_btn)
    assert np.array_equal((data1+data2)/2, integration_widget.img_widget.img_data)


def test_load_image_with_manual_input_file_name(
    integration_widget, dioptas_model, image_controller
):
    file_name = os.path.join(unittest_data_path, "LaB6_40keV_MarCCD.tif")
    integration_widget.img_filename_txt.setText(file_name)
    image_controller.filename_txt_changed()  # need to manually trigger the text change
    old_data = np.copy(dioptas_model.img_data)

    file_name = os.path.join(unittest_data_path, "CeO2_Pilatus1M.tif")
    integration_widget.img_filename_txt.setText(file_name)
    image_controller.filename_txt_changed()
    new_data = dioptas_model.img_data

    assert not np.array_equal(old_data, new_data)


def test_changing_cake_integral_width(
    integration_widget, dioptas_model, image_controller
):
    load_pilatus1M_image_and_calibration(dioptas_model)
    click_button(integration_widget.integration_image_widget.mode_btn)
    image_controller.img_mouse_click(100, 300)

    x = integration_widget.cake_widget.cake_integral_item.xData
    integration_widget.integration_control_widget.integration_options_widget.cake_integral_width_sb.setValue(
        3
    )
    image_controller.img_mouse_click(100, 300)

    assert not np.array_equal(
        x, integration_widget.cake_widget.cake_integral_item.xData
    )


def test_clicking_cake_image(integration_widget, dioptas_model, image_controller):
    load_pilatus1M_image_and_calibration(dioptas_model)
    click_button(integration_widget.integration_image_widget.mode_btn)
    integration_widget.integration_image_widget.cake_view.mouse_left_clicked.emit(
        30, 40
    )


def test_click_image_sends_tth_changed_signal(
    integration_widget, dioptas_model, image_controller
):
    load_pilatus1M_image_and_calibration(dioptas_model)
    dioptas_model.clicked_tth_changed.emit = MagicMock()
    integration_widget.integration_image_widget.img_view.mouse_left_clicked.emit(
        600, 400
    )
    dioptas_model.clicked_tth_changed.emit.assert_called()


def test_click_cake_sends_tth_changed_signal(
    integration_widget, dioptas_model, image_controller
):
    load_pilatus1M_image_and_calibration(dioptas_model)
    dioptas_model.clicked_tth_changed.emit = MagicMock()
    click_button(integration_widget.integration_image_widget.mode_btn)
    integration_widget.integration_image_widget.cake_view.mouse_left_clicked.emit(
        1100, 50
    )
    dioptas_model.clicked_tth_changed.emit.assert_called_once_with(
        get_partial_value(dioptas_model.cake_tth, 1100 - 0.5)
    )


def test_clicked_tth_changed(integration_widget, dioptas_model, image_controller):
    load_pilatus1M_image_and_calibration(dioptas_model)
    integration_widget.integration_image_widget.img_view.mouse_left_clicked.emit(
        600, 400
    )
    before_circle_data = (
        integration_widget.integration_image_widget.img_view.circle_plot_items[
            0
        ].getData()
    )
    dioptas_model.clicked_tth_changed.emit(10)
    after_circle_data = (
        integration_widget.integration_image_widget.img_view.circle_plot_items[
            0
        ].getData()
    )
    assert not np.array_equal(before_circle_data, after_circle_data)


def test_circle_scatter_is_activated_correctly(
    integration_widget, dioptas_model, image_controller
):
    dioptas_model.clicked_tth_changed.emit(10)
    assert not (
        integration_widget.integration_image_widget.img_view.circle_plot_items[0]
        in integration_widget.img_widget.img_view_box.addedItems
    )

    load_pilatus1M_image_and_calibration(dioptas_model)
    dioptas_model.clicked_tth_changed.emit(10)
    assert (
        integration_widget.integration_image_widget.img_view.circle_plot_items[0]
        in integration_widget.img_widget.img_view_box.addedItems
    )


def test_loading_series_karabo_file_shows_correct_gui(
    integration_widget, dioptas_model, image_controller
):
    from dioptas.model.loader.KaraboLoader import extra_data_installed

    if not extra_data_installed:
        return
    filename = os.path.join(unittest_data_path, "karabo_epix.h5")
    file_widget = (
        integration_widget.integration_control_widget.img_control_widget.file_widget
    )
    integration_widget.show()
    assert not file_widget.step_series_widget.isVisible()
    dioptas_model.img_model.load(filename)
    assert file_widget.step_series_widget.isVisible()


def test_fileinfo_and_move_button_visibility(
    integration_widget, dioptas_model, image_controller
):
    filename = os.path.join(unittest_data_path, "image_001.tif")
    integration_widget.show()
    dioptas_model.img_model.load(filename)
    assert not integration_widget.file_info_btn.isVisible()

    filename = os.path.join(unittest_data_path, "TransferCorrection", "original.tif")
    dioptas_model.img_model.load(filename)
    assert integration_widget.file_info_btn.isVisible()


def test_sources_for_hdf5_files(integration_widget, dioptas_model, image_controller):
    file_widget = (
        integration_widget.integration_control_widget.img_control_widget.file_widget
    )
    integration_widget.show()
    assert not file_widget.sources_widget.isVisible()

    # load file with different sources
    filename = os.path.join(unittest_data_path, "hdf5_dataset", "ma4500_demoh5.h5")
    dioptas_model.img_model.load(filename)
    assert file_widget.sources_widget.isVisible()

    # load file without sources
    filename = os.path.join(unittest_data_path, "image_001.tif")
    dioptas_model.img_model.load(filename)
    assert not file_widget.sources_widget.isVisible()


def test_sources_are_updated_in_sources_combobox(
    integration_widget, dioptas_model, image_controller
):
    file_widget = (
        integration_widget.integration_control_widget.img_control_widget.file_widget
    )
    filename = os.path.join(unittest_data_path, "hdf5_dataset", "ma4500_demoh5.h5")
    dioptas_model.img_model.load(filename)

    assert file_widget.sources_cb.count() > 0
    assert file_widget.sources_cb.count() == len(dioptas_model.img_model.sources)

    file_widget.sources_cb.setCurrentIndex(2)
    assert file_widget.sources_cb.currentText() == dioptas_model.img_model.sources[2]


def load_pilatus1M_image_and_calibration(dioptas_model: DioptasModel):
    file_name = os.path.join(unittest_data_path, "CeO2_Pilatus1M.tif")
    dioptas_model.img_model.load(file_name)
    calibration_file_name = os.path.join(unittest_data_path, "CeO2_Pilatus1M.poni")
    dioptas_model.calibration_model.load(calibration_file_name)



# SPDX-License-Identifier: MIT

from ..utility import QtTest
import os
import gc

from mock import MagicMock
import numpy as np

from ...controller.integration import PhaseInPatternController, PatternController
from ...model.DioptasModel import DioptasModel
from ...widgets.integration import IntegrationWidget


unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, "../data")
jcpds_path = os.path.join(data_path, "jcpds")


class PhaseInPatternControllerTest(QtTest):
    def setUp(self) -> None:
        self.model = DioptasModel()
        self.model.calibration_model.is_calibrated = True
        self.model.calibration_model.pattern_geometry.wavelength = 0.31e-10
        self.model.calibration_model.integrate_1d = MagicMock(
            return_value=(
                self.model.calibration_model.tth,
                self.model.calibration_model.int,
            )
        )
        self.widget = IntegrationWidget()

        self.controller = PhaseInPatternController(
            self.widget.pattern_widget, self.model
        )
        self.pattern_controller = PatternController(self.widget, self.model)
        self.model.pattern_model.load_pattern(os.path.join(data_path, "pattern_001.xy"))

    def tearDown(self) -> None:
        del self.controller
        del self.widget
        self.model.delete_configurations()
        del self.model
        gc.collect()

    # Utility Functions
    #######################
    def load_phase(self, filename):
        self.model.phase_model.add_jcpds(os.path.join(jcpds_path, filename))

    def load_phases(self):
        self.load_phase("ar.jcpds")
        self.load_phase("ag.jcpds")
        self.load_phase("au_Anderson.jcpds")
        self.load_phase("mo.jcpds")
        self.load_phase("pt.jcpds")
        self.load_phase("re.jcpds")

    # Tests
    #######################
    def test_loading_a_phase(self):
        self.assertEqual(len(self.widget.pattern_widget.phases), 0)
        self.load_phase("ar.jcpds")

        self.assertEqual(len(self.widget.pattern_widget.phases), 1)

    def test_remove_phase(self):
        self.load_phases()
        self.assertEqual(len(self.widget.pattern_widget.phases), 6)
        self.model.phase_model.del_phase(3)
        self.assertEqual(len(self.widget.pattern_widget.phases), 5)

    def test_changing_pressure(self):
        self.load_phase("ar.jcpds")
        first_line_position = (
            self.widget.pattern_widget.phases[0].line_items[0].getData()[0][0]
        )
        self.model.phase_model.set_pressure(0, 4)
        self.assertNotEqual(
            first_line_position,
            self.widget.pattern_widget.phases[0].line_items[0].getData()[0][0],
        )

    def test_changing_temperature_and_pressure(self):
        self.load_phase("pt.jcpds")
        self.model.phase_model.set_pressure(0, 100)
        first_line_position = (
            self.widget.pattern_widget.phases[0].line_items[0].getData()[0][0]
        )
        self.model.phase_model.set_temperature(0, 3000)
        self.assertNotEqual(
            first_line_position,
            self.widget.pattern_widget.phases[0].line_items[0].getData()[0][0],
        )

    def test_changing_color(self):
        self.load_phase("pt.jcpds")
        green_value = (
            self.widget.pattern_widget.phases[0]
            .line_items[0]
            .opts["pen"]
            .color()
            .green()
        )
        self.model.phase_model.set_color(0, (230, 22, 0))
        self.assertNotEqual(
            green_value,
            self.widget.pattern_widget.phases[0]
            .line_items[0]
            .opts["pen"]
            .color()
            .green(),
        )

    def test_auto_scaling_of_lines_in_pattern_view(self):
        pattern_view = self.widget.pattern_widget

        pattern_view_range = pattern_view.view_box.viewRange()
        pattern_y = pattern_view.plot_item.getData()[1]
        expected_maximum_height = np.max(pattern_y) - pattern_view_range[1][0]

        self.load_phase("au_Anderson.jcpds")
        phase_plot = pattern_view.phases[0]
        line_heights = []
        for line in phase_plot.line_items:
            line_data = line.getData()
            height = line_data[1][1] - line_data[1][0]
            line_heights.append(height)

        self.assertAlmostEqual(expected_maximum_height, np.max(line_heights))

        pattern_view_range = pattern_view.view_box.viewRange()
        pattern_y = pattern_view.plot_item.getData()[1]
        expected_maximum_height = np.max(pattern_y) - pattern_view_range[1][0]

        self.assertAlmostEqual(expected_maximum_height, np.max(line_heights))

    def test_line_height_in_pattern_view_after_zooming(self):
        pattern_view = self.widget.pattern_widget
        self.load_phase("au_Anderson.jcpds")

        pattern_view.view_box.setRange(xRange=[17, 30])
        pattern_view.emit_sig_range_changed()

        phase_plot = pattern_view.phases[0]
        line_heights = []
        for line in phase_plot.line_items:
            line_data = line.getData()
            if (line_data[0][0] > 17) and (line_data[0][1] < 30):
                height = line_data[1][1] - line_data[1][0]
                line_heights.append(height)

        pattern_view_range = pattern_view.view_box.viewRange()
        pattern_x, pattern_y = pattern_view.plot_item.getData()
        pattern_y_max_in_range = np.max(
            pattern_y[
                (pattern_x > pattern_view_range[0][0])
                & (pattern_x < pattern_view_range[0][1])
            ]
        )
        expected_maximum_height = pattern_y_max_in_range - pattern_view_range[1][0]

        self.assertAlmostEqual(expected_maximum_height, np.max(line_heights))

    def test_add_reflection(self):
        pattern_view = self.widget.pattern_widget
        self.load_phase("au_Anderson.jcpds")
        num_line_items = len(pattern_view.phases[0].line_items)
        self.model.phase_model.add_reflection(0)
        self.assertEqual(len(pattern_view.phases[0].line_items), num_line_items + 1)

    def test_delete_reflection(self):
        pattern_view = self.widget.pattern_widget
        self.load_phase("au_Anderson.jcpds")
        num_line_items = len(pattern_view.phases[0].line_items)
        self.model.phase_model.delete_reflection(0, 0)
        self.assertEqual(len(pattern_view.phases[0].line_items), num_line_items - 1)

    def test_reload_phase(self):
        self.load_phase("au_Anderson.jcpds")
        self.model.phase_model.reload(0)

    def test_reload_phase_after_deleting_a_reflection(self):
        self.load_phase("au_Anderson.jcpds")
        self.model.phase_model.delete_reflection(0, 0)
        self.model.phase_model.reload(0)



# SPDX-License-Identifier: MIT

import unittest
from qtpy import QtWidgets, QtCore
from qtpy.QtTest import QTest
import os
import shutil

unittest_data_path = os.path.join(os.path.dirname(__file__), 'data')


class QtTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.app = QtWidgets.QApplication.instance()
        if cls.app is None:
            cls.app = QtWidgets.QApplication([])


def delete_if_exists(data_path):
    if os.path.exists(data_path):
        os.remove(data_path)


def delete_folder_if_exists(data_path):
    if os.path.exists(data_path):
        shutil.rmtree(data_path)


def click_button(widget):
    QTest.mouseClick(widget, QtCore.Qt.LeftButton)


def click_checkbox(checkbox_widget):
    QTest.mouseClick(checkbox_widget, QtCore.Qt.LeftButton,
                     pos=QtCore.QPoint(2, int(checkbox_widget.height() / 2.0)))


def enter_value_into_text_field(text_field, value):
    text_field.setText('')
    QTest.keyClicks(text_field, str(value))
    QTest.keyPress(text_field, QtCore.Qt.Key_Enter)
    QtWidgets.QApplication.processEvents()


class MockMouseEvent:
    def __init__(self, key=None, diff=None):
        self.key_value = key
        self.diff = diff

        class TestCoord:
            def x(self):
                return 100

            def y(self):
                return 100

        self.coord = TestCoord()

    def key(self):
        return self.key_value

    def x(self):
        return self.diff

    def angleDelta(self):
        return self.coord

    def modifiers(self):
        return QtCore.Qt.CoverWindow

    def button(self):
        return QtCore.Qt.CoverWindow



# SPDX-License-Identifier: MIT

import logging
import os
import struct
from collections import deque
from concurrent.futures import ThreadPoolExecutor

import h5py
import hdf5plugin
import numpy as np

try:
    import bitshuffle

    HAS_BITSHUFFLE = True
except ImportError:
    HAS_BITSHUFFLE = False

logger = logging.getLogger(__name__)

BITSHUFFLE_FILTER_ID = 32008
_NUM_CPUS = max((os.cpu_count() or 4) - 1, 1)


def _decompress_bitshuffle_lz4(raw_bytes, shape, dtype):
    """Decompress a single bitshuffle-LZ4 compressed chunk and return as
    contiguous float64 with vertical flip applied.

    This runs inside a thread pool worker so the flip + type conversion
    happen in parallel across frames.
    """
    block_size_bytes = struct.unpack(">I", raw_bytes[8:12])[0]
    block_size = block_size_bytes // np.dtype(dtype).itemsize
    buf = np.frombuffer(raw_bytes, dtype=np.uint8, offset=12)
    frame = bitshuffle.decompress_lz4(buf, shape, dtype, block_size)
    return np.ascontiguousarray(frame[::-1], dtype=np.float64)


def _is_bitshuffle_compressed(dataset):
    """Check if an HDF5 dataset uses bitshuffle compression."""
    if not HAS_BITSHUFFLE:
        return False
    try:
        plist = dataset.id.get_create_plist()
        n_filters = plist.get_nfilters()
        for i in range(n_filters):
            filter_info = plist.get_filter(i)
            if filter_info[0] == BITSHUFFLE_FILTER_ID:
                return True
    except Exception:
        pass
    return False


class Hdf5Image:
    def __init__(self, filename):
        """
        Loads an Hdf5 image produced by ESRF
        :param filename: path to the hdf5 file to be loaded
        """

        self.filename = filename
        self.f = h5py.File(filename, "r")
        self.image_sources = find_image_sources(self.f)

        self.dataset = self.f[self.image_sources[0]]
        self.series_max = self.dataset.shape[0]
        self._is_bitshuffle = _is_bitshuffle_compressed(self.dataset)

    def get_image(self, ind):
        return self.dataset[ind][::-1]

    def select_source(self, source):
        self.dataset = self.f[source]
        self.series_max = self.dataset.shape[0]
        self._is_bitshuffle = _is_bitshuffle_compressed(self.dataset)

    def gen_frames(self, n_frames=None, decomp_workers=_NUM_CPUS):
        """Generator yielding all frames as contiguous float64 arrays.

        For bitshuffle-compressed datasets, decompression, vertical flip, and
        float64 conversion all happen in parallel worker threads.  For other
        datasets, frames are read sequentially through h5py.

        :param n_frames: number of frames to yield (default: all)
        :param decomp_workers: number of parallel decompression threads
        """
        n = self.series_max if n_frames is None else min(n_frames, self.series_max)

        if self._is_bitshuffle:
            yield from self._gen_frames_parallel(n, decomp_workers)
        else:
            for i in range(n):
                yield np.ascontiguousarray(
                    self.dataset[i][::-1], dtype=np.float64
                )

    def _gen_frames_parallel(self, n_frames, decomp_workers=_NUM_CPUS):
        """Read raw chunks and decompress bitshuffle-LZ4 in parallel threads.

        Each worker decompresses a chunk, applies the vertical flip, and
        converts to contiguous float64.  A sliding window of at most
        *decomp_workers* in-flight futures limits memory usage and CPU
        contention with downstream integration threads.
        """
        ds = self.dataset
        dtype = ds.dtype
        chunk_shape = ds.chunks
        frame_shape = chunk_shape[1:]

        # Verify each frame is a single chunk (required for read_direct_chunk)
        if chunk_shape[1:] != ds.shape[1:]:
            raise ValueError(
                f"Spatial chunking detected: chunk_shape={chunk_shape}, "
                f"dataset_shape={ds.shape}. Parallel bitshuffle "
                f"decompression requires one chunk per frame."
            )

        with ThreadPoolExecutor(max_workers=decomp_workers) as pool:
            pending = deque()
            idx = 0

            # Fill initial window
            while idx < n_frames and len(pending) < decomp_workers:
                raw = bytes(ds.id.read_direct_chunk((idx, 0, 0))[1])
                pending.append(
                    pool.submit(
                        _decompress_bitshuffle_lz4, raw, frame_shape, dtype
                    )
                )
                idx += 1

            # Yield results and refill window one-for-one
            while pending:
                yield pending.popleft().result()
                if idx < n_frames:
                    raw = bytes(ds.id.read_direct_chunk((idx, 0, 0))[1])
                    pending.append(
                        pool.submit(
                            _decompress_bitshuffle_lz4, raw, frame_shape, dtype
                        )
                    )
                    idx += 1


def find_image_sources(hd5_file):
    image_paths = []

    def traverse_groups(group, parent_path=""):
        if isinstance(group, h5py.Dataset):
            if len(group.shape) >= 3:
                image_paths.append(parent_path)
        else:  # node is a group
            for key in group.keys():
                traverse_groups(group[key], parent_path + "/" + key)

    traverse_groups(hd5_file)

    return image_paths



# SPDX-License-Identifier: MIT

import unittest
from mock import MagicMock, patch
import os
import gc
from pyFAI import detectors

import numpy as np

from qtpy import QtWidgets, QtCore
from qtpy.QtTest import QTest

from ..utility import (
    QtTest,
    unittest_data_path,
    click_button,
    click_checkbox,
    delete_if_exists,
)
from ...model.DioptasModel import DioptasModel
from ...controller.CalibrationController import (
    CalibrationController,
    get_available_detectors,
)
from ...widgets.CalibrationWidget import CalibrationWidget

class TestCalibrationController(QtTest):
    def setUp(self):
        # Mocking the functions which will block the unittest for some reason...
        # Use patch to ensure proper cleanup
        self.processEvents_patcher = patch.object(QtWidgets.QApplication, 'processEvents', MagicMock())
        self.setValue_patcher = patch.object(QtWidgets.QProgressDialog, 'setValue', MagicMock())
        self.processEvents_patcher.start()
        self.setValue_patcher.start()

        self.model = DioptasModel()

        self.widget = CalibrationWidget()
        self.controller = CalibrationController(
            widget=self.widget, dioptas_model=self.model
        )

    def tearDown(self):
        # Stop the patchers to restore original behavior
        self.processEvents_patcher.stop()
        self.setValue_patcher.stop()

        delete_if_exists(os.path.join(unittest_data_path, "detector_with_spline.h5"))

        # Clean up widgets
        self.widget.close()
        self.widget.deleteLater()
        del self.widget
        del self.controller
        del self.model
        gc.collect()

    def mock_integrate_functions(self):
        self.model.calibration_model.integrate_1d = MagicMock(
            return_value=([np.linspace(0, 100), np.linspace(0, 100)])
        )
        self.model.calibration_model.integrate_2d = MagicMock()

    def test_load_detector(self):
        detector_names, detector_classes = get_available_detectors()
        det_ind = 9
        self.widget.detectors_cb.setCurrentIndex(
            det_ind + 3
        )  # +3 since there is also the custom element at 0
        # and 2 separators
        self.assertIsInstance(
            self.model.calibration_model.detector, detector_classes[det_ind]
        )

        detector_gb = (
            self.widget.calibration_control_widget.calibration_parameters_widget.detector_gb
        )
        self.assertAlmostEqual(
            float(detector_gb.pixel_width_txt.text()) * 1e-6,
            self.model.calibration_model.orig_pixel1,
        )
        self.assertAlmostEqual(
            float(detector_gb.pixel_height_txt.text()) * 1e-6,
            self.model.calibration_model.orig_pixel2,
        )

        self.assertFalse(detector_gb.pixel_width_txt.isEnabled())
        self.assertFalse(detector_gb.pixel_width_txt.isEnabled())

        self.widget.detectors_cb.setCurrentIndex(0)
        self.assertNotIsInstance(
            self.model.calibration_model.detector, detector_classes[det_ind]
        )
        self.assertAlmostEqual(
            float(detector_gb.pixel_width_txt.text()) * 1e-6,
            self.model.calibration_model.orig_pixel1,
        )
        self.assertAlmostEqual(
            float(detector_gb.pixel_height_txt.text()) * 1e-6,
            self.model.calibration_model.orig_pixel2,
        )
        self.assertTrue(detector_gb.pixel_width_txt.isEnabled())
        self.assertTrue(detector_gb.pixel_width_txt.isEnabled())

    def test_load_detector_transform_and_reset(self):
        QtWidgets.QFileDialog.getOpenFileName = MagicMock(
            return_value=os.path.join(unittest_data_path, "CeO2_Pilatus1M.poni")
        )
        QTest.mouseClick(self.widget.load_calibration_btn, QtCore.Qt.LeftButton)

        detector_gb = (
            self.widget.calibration_control_widget.calibration_parameters_widget.detector_gb
        )
        detector_gb.detector_cb.setCurrentIndex(
            detector_gb.detector_cb.findText("Pilatus CdTe 1M")
        )

        QtWidgets.QFileDialog.getOpenFileName = MagicMock(
            return_value=os.path.join(unittest_data_path, "CeO2_Pilatus1M.tif")
        )
        QTest.mouseClick(self.widget.load_img_btn, QtCore.Qt.LeftButton)

        QTest.mouseClick(self.widget.rotate_m90_btn, QtCore.Qt.LeftButton)
        QTest.mouseClick(self.widget.rotate_m90_btn, QtCore.Qt.LeftButton)

        detector_gb.detector_cb.setCurrentIndex(
            detector_gb.detector_cb.findText("Custom")
        )

        QTest.mouseClick(self.widget.rotate_m90_btn, QtCore.Qt.LeftButton)

    def test_load_detector_from_file(self):
        QtWidgets.QFileDialog.getOpenFileName = MagicMock(
            return_value=os.path.join(unittest_data_path, "detector.h5")
        )
        click_button(self.widget.detector_load_btn)

        self.assertAlmostEqual(self.model.calibration_model.orig_pixel1, 100e-6)

        self.assertTrue(self.widget.detector_reset_btn.isEnabled())
        self.widget.show()  # to check visibility
        self.assertFalse(self.widget.detectors_cb.isVisible())
        self.assertTrue(self.widget.detector_name_lbl.isVisible())
        self.assertEqual(self.widget.detector_name_lbl.text(), "detector.h5")
        detector_gb = (
            self.widget.calibration_control_widget.calibration_parameters_widget.detector_gb
        )
        self.assertFalse(detector_gb.pixel_width_txt.isEnabled())
        self.assertFalse(detector_gb.pixel_height_txt.isEnabled())
        self.assertAlmostEqual(self.widget.get_pixel_size()[0], 100e-6)

        click_button(self.widget.detector_reset_btn)
        self.assertTrue(self.widget.detectors_cb.isVisible())
        self.assertFalse(self.widget.detector_name_lbl.isVisible())
        self.assertEqual(self.widget.detectors_cb.currentIndex(), 0)
        self.assertTrue(detector_gb.pixel_width_txt.isEnabled())
        self.assertTrue(detector_gb.pixel_height_txt.isEnabled())

    def test_load_detector_with_distortion(self):
        # create detector and save it
        spline_detector = detectors.Detector()
        spline_detector.splinefile = os.path.join(
            unittest_data_path, "distortion", "f4mnew.spline"
        )
        spline_detector.save(
            os.path.join(unittest_data_path, "detector_with_spline.h5")
        )

        QtWidgets.QFileDialog.getOpenFileName = MagicMock(
            return_value=os.path.join(unittest_data_path, "detector_with_spline.h5")
        )
        click_button(self.widget.detector_load_btn)

        self.assertEqual(self.widget.spline_filename_txt.text(), "from Detector")

    def test_load_detector_and_image_with_different_dimension(self):
        # set the detector
        self.widget.detectors_cb.setCurrentIndex(
            self.widget.detectors_cb.findText("Pilatus CdTe 1M")
        )

        # load calibration
        QtWidgets.QFileDialog.getOpenFileName = MagicMock(
            return_value=os.path.join(unittest_data_path, "CeO2_Pilatus1M.poni")
        )
        QTest.mouseClick(self.widget.load_calibration_btn, QtCore.Qt.LeftButton)

        # load image file with different dimension than detector, which should automatically
        # reset the detector to a custom detector
        QtWidgets.QMessageBox.critical = MagicMock()
        QtWidgets.QFileDialog.getOpenFileName = MagicMock(
            return_value=os.path.join(unittest_data_path, "LaB6_40keV_MarCCD.tif")
        )
        click_button(self.widget.load_img_btn)

        QtWidgets.QMessageBox.critical.assert_called_once()
        self.assertEqual(self.widget.detectors_cb.currentText(), "Custom")

    def test_automatic_calibration(self):
        self.mock_integrate_functions()
        QtWidgets.QFileDialog.getOpenFileName = MagicMock(
            return_value=os.path.join(unittest_data_path, "LaB6_40keV_MarCCD.tif")
        )
        QTest.mouseClick(self.widget.load_img_btn, QtCore.Qt.LeftButton)
        self.controller.search_peaks(1179.6, 1129.4)
        self.controller.search_peaks(1268.5, 1119.8)
        self.controller.widget.sv_wavelength_txt.setText("0.31")
        self.controller.widget.sv_distance_txt.setText("200")
        self.controller.widget.set_pixel_size(79e-6, 79e-6)
        calibrant_index = self.widget.calibrant_cb.findText("LaB6")
        self.controller.widget.calibrant_cb.setCurrentIndex(calibrant_index)
        #
        QTest.mouseClick(self.widget.calibrate_btn, QtCore.Qt.LeftButton)
        self.app.processEvents()
        self.model.calibration_model.integrate_1d.assert_called_once()
        self.model.calibration_model.integrate_2d.assert_called_once()
        # Progress dialog should be updated during calibration
        self.assertGreater(QtWidgets.QProgressDialog.setValue.call_count, 0)

        calibration_parameter = (
            self.model.calibration_model.get_calibration_parameter()[0]
        )
        self.assertAlmostEqual(calibration_parameter["dist"], 0.1968, places=3)

    def test_searching_peaks_automatic_increase(self):
        self.mock_integrate_functions()
        QtWidgets.QFileDialog.getOpenFileName = MagicMock(
            return_value=os.path.join(unittest_data_path, "LaB6_40keV_MarCCD.tif")
        )
        QTest.mouseClick(self.widget.load_img_btn, QtCore.Qt.LeftButton)
        self.controller.search_peaks(1179.6, 1129.4)
        self.controller.search_peaks(1268.5, 1119.8)
        self.assertEqual(3, self.widget.peak_num_sb.value())

    def test_searching_peaks_manual_index_assignment(self):
        self.mock_integrate_functions()
        QtWidgets.QFileDialog.getOpenFileName = MagicMock(
            return_value=os.path.join(unittest_data_path, "LaB6_40keV_MarCCD.tif")
        )
        QTest.mouseClick(self.widget.load_img_btn, QtCore.Qt.LeftButton)

        click_checkbox(self.widget.automatic_peak_num_inc_cb)
        self.assertFalse(self.widget.automatic_peak_num_inc_cb.isChecked())

        self.controller.search_peaks(1179.6, 1129.4)
        self.controller.search_peaks(1268.5, 1119.8)
        self.assertEqual(1, self.widget.peak_num_sb.value())

    def test_splines(self):
        self.mock_integrate_functions()
        QtWidgets.QFileDialog.getOpenFileName = MagicMock(
            return_value=os.path.join(unittest_data_path, "distortion", "f4mnew.spline")
        )
        click_button(self.widget.load_spline_btn)

        self.assertIsNotNone(self.model.calibration_model.distortion_spline_filename)
        self.assertEqual(self.widget.spline_filename_txt.text(), "f4mnew.spline")
        self.assertTrue(self.widget.spline_reset_btn.isEnabled())
        #
        click_button(self.widget.spline_reset_btn)
        self.assertIsNone(self.model.calibration_model.distortion_spline_filename)
        self.assertEqual(self.widget.spline_filename_txt.text(), "None")

    def test_loading_and_saving_of_calibration_files(self):
        self.mock_integrate_functions()
        QtWidgets.QFileDialog.getOpenFileName = MagicMock(
            return_value=os.path.join(unittest_data_path, "LaB6_40keV_MarCCD.poni")
        )
        QTest.mouseClick(self.widget.load_calibration_btn, QtCore.Qt.LeftButton)
        QtWidgets.QFileDialog.getSaveFileName = MagicMock(
            return_value=os.path.join(unittest_data_path, "calibration.poni")
        )
        QTest.mouseClick(self.widget.save_calibration_btn, QtCore.Qt.LeftButton)
        self.assertTrue(
            os.path.exists(os.path.join(unittest_data_path, "calibration.poni"))
        )
        os.remove(os.path.join(unittest_data_path, "calibration.poni"))

    def test_selecting_configuration_updates_parameter_display(self):
        self.mock_integrate_functions()
        calibration1 = {
            "dist": 0.2,
            "poni1": 0.08,
            "poni2": 0.081,
            "rot1": 0.0043,
            "rot2": 0.002,
            "rot3": 0.001,
            "pixel1": 7.9e-5,
            "pixel2": 7.9e-5,
            "wavelength": 0.3344,
            "polarization_factor": 0.99,
        }
        calibration2 = {
            "dist": 0.3,
            "poni1": 0.04,
            "poni2": 0.021,
            "rot1": 0.0053,
            "rot2": 0.002,
            "rot3": 0.0013,
            "pixel1": 7.4e-5,
            "pixel2": 7.6e-5,
            "wavelength": 0.31,
            "polarization_factor": 0.98,
        }

        self.model.calibration_model.set_pyFAI(calibration1)
        self.model.add_configuration()
        self.model.calibration_model.set_pyFAI(calibration2)

        self.model.select_configuration(0)

        model_calibration = self.model.configurations[
            0
        ].calibration_model.pattern_geometry.getPyFAI()
        del model_calibration["detector"]
        for key in ("splineFile", "splinefile", "max_shape", "orientation"):
            model_calibration.pop(key, None)

        current_displayed_calibration = self.widget.get_pyFAI_parameter()
        del current_displayed_calibration["polarization_factor"]
        self.assertEqual(model_calibration, current_displayed_calibration)

        self.model.select_configuration(1)
        model_calibration = self.model.configurations[
            1
        ].calibration_model.pattern_geometry.getPyFAI()
        del model_calibration["detector"]
        for key in ("splineFile", "splinefile", "max_shape", "orientation"):
            model_calibration.pop(key, None)
        current_displayed_calibration = self.widget.get_pyFAI_parameter()
        del current_displayed_calibration["polarization_factor"]

        self.assertEqual(model_calibration, current_displayed_calibration)

    @unittest.skip("Does not work for unknown reasons")
    def test_calibrant_with_small_set_of_d_spacings(self):
        self.mock_integrate_functions()
        QtWidgets.QFileDialog.getOpenFileName = MagicMock(
            return_value=os.path.join(unittest_data_path, "LaB6_40keV_MarCCD.tif")
        )
        QTest.mouseClick(self.widget.load_img_btn, QtCore.Qt.LeftButton)
        self.controller.search_peaks(1179.6, 1129.4)
        self.controller.search_peaks(1268.5, 1119.8)
        calibrant_index = self.widget.calibrant_cb.findText("CuO")
        self.controller.widget.calibrant_cb.setCurrentIndex(calibrant_index)
        QtWidgets.QMessageBox.critical = MagicMock()
        click_button(self.widget.calibrate_btn)
        QtWidgets.QMessageBox.critical.assert_called_once()

    def test_loading_calibration_without_an_image_before(self):
        QtWidgets.QFileDialog.getOpenFileName = MagicMock(
            return_value=os.path.join(unittest_data_path, "LaB6_40keV_MarCCD.poni")
        )
        QTest.mouseClick(self.widget.load_calibration_btn, QtCore.Qt.LeftButton)
        self.widget.get_pyFAI_parameter()  # would cause error if GUI not updated
        self.widget.get_fit2d_parameter()  # would cause error if GUI not updated

    def test_detector_rotation_does_not_emit_reset_detector_signal(self):
        QtWidgets.QFileDialog.getOpenFileName = MagicMock(
            return_value=os.path.join(unittest_data_path, "LaB6_40keV_MarCCD.tif")
        )
        click_button(self.widget.load_img_btn)
        self.model.calibration_model.detector_reset.emit = MagicMock()

        click_button(self.widget.rotate_m90_btn)
        self.model.calibration_model.detector_reset.emit.assert_not_called()


def test_click_calibrate_without_defined_points(
    calibration_controller, calibration_model
):
    QtWidgets.QMessageBox.critical = MagicMock()
    calibration_model.calibrate = MagicMock()
    click_button(calibration_controller.widget.calibrate_btn)
    assert QtWidgets.QMessageBox.critical.called
    assert not calibration_model.calibrate.called


def test_click_refine_without_defined_points(calibration_controller, calibration_model):
    QtWidgets.QMessageBox.critical = MagicMock()
    click_button(calibration_controller.widget.refine_btn)
    assert QtWidgets.QMessageBox.critical.called



# SPDX-License-Identifier: MIT
"""Test ColormapPopup widget"""

import numpy as np
import pytest
from qtpy import QtCore, QtWidgets
from qtpy.QtTest import QSignalSpy, QTest

import pyqtgraph.graphicsItems.GradientEditorItem

from ...widgets.plot_widgets.ColormapPopup import ColormapPopup


def testRange(qWidgetFactory):
    """Test getRange, setRange and sigRangeChanged"""
    colormapPopup = qWidgetFactory(ColormapPopup)
    assert colormapPopup.getRange() == (1, 1)

    signalSpy = QSignalSpy(colormapPopup.sigRangeChanged)

    colormapPopup.setRange(100, 1000)
    assert len(signalSpy) == 1
    assert signalSpy[0] == [100, 1000]
    assert colormapPopup.getRange() == (100, 1000)

    colormapPopup.setRange(2000, 1000)
    assert len(signalSpy) == 2
    assert signalSpy[1] == [1000, 2000]
    assert colormapPopup.getRange() == (1000, 2000)


def testCurrentGradient(qWidgetFactory):
    """Test getCurrentGradient, setCurrentGradient and sigCurrentGradientChanged"""
    colormapPopup = qWidgetFactory(ColormapPopup)
    for firstName, firstGradient in pyqtgraph.graphicsItems.GradientEditorItem.Gradients.items():
        break
    gradient = colormapPopup.getCurrentGradient()
    assert gradient == firstGradient
    assert colormapPopup._gradientComboBox.currentText() == firstName.capitalize()
 
    signalSpy = QSignalSpy(colormapPopup.sigCurrentGradientChanged)
    viridisGradient = pyqtgraph.graphicsItems.GradientEditorItem.Gradients['viridis']
    colormapPopup.setCurrentGradient(viridisGradient)
    gradient = colormapPopup.getCurrentGradient()
    assert gradient == viridisGradient
    assert colormapPopup._gradientComboBox.currentText() == 'Viridis'
    assert len(signalSpy) == 1
    assert signalSpy[0] == [viridisGradient]


def testCustomGradient(qWidgetFactory):
    """Test setCurrentGradient with a custom gradient"""
    colormapPopup = qWidgetFactory(ColormapPopup)
    signalSpy = QSignalSpy(colormapPopup.sigCurrentGradientChanged)

    customGradient = {
        'mode': 'rgb',
        'ticks': [(0.0, (0, 0, 0, 255)), (1.0, (0, 0, 0, 255))],
    }
    colormapPopup.setCurrentGradient(customGradient)
    gradient = colormapPopup.getCurrentGradient()
    assert gradient == customGradient
    assert colormapPopup._gradientComboBox.currentText() == 'Custom'
    assert len(signalSpy) == 1
    assert signalSpy[0] == [customGradient]

    customGradient2 = {
        'mode': 'rgb',
        'ticks': [(0.0, (255, 255, 255, 255)), (1.0, (255, 255, 255, 255))],
    }
    colormapPopup.setCurrentGradient(customGradient2)
    gradient = colormapPopup.getCurrentGradient()
    assert gradient == customGradient2
    assert colormapPopup._gradientComboBox.currentText() == 'Custom'
    assert len(signalSpy) == 2
    assert signalSpy[1] == [customGradient2]


@pytest.mark.parametrize("normalization", ["log", "sqrt"])
def testCurrentNormalization(qWidgetFactory, normalization):
    """Test getCurrentNormalization, setCurrentNormalization and sigCurrentNormalizationChanged"""
    colormapPopup = qWidgetFactory(ColormapPopup)

    default_normalization = colormapPopup.getCurrentNormalization()
    assert default_normalization == "linear"
    assert colormapPopup._normalizationComboBox.currentText() == "Linear"
    assert colormapPopup._normalizationComboBox.currentData() == "linear"

    signalSpy = QSignalSpy(colormapPopup.sigCurrentNormalizationChanged)
    colormapPopup.setCurrentNormalization(normalization)
    returned_normalization = colormapPopup.getCurrentNormalization()
    assert returned_normalization == normalization
    assert colormapPopup._normalizationComboBox.currentData() == normalization
    assert len(signalSpy) == 1
    assert signalSpy[0] == [normalization]


def testResetMode(qWidgetFactory):
    """Test reset range and changing reset mode"""
    colormapPopup = qWidgetFactory(ColormapPopup)
    colormapPopup.setData(np.arange(101))

    buttons = colormapPopup._resetButtonGroup.buttons()
    percentileButton, minmaxButton, mean3stdButton = buttons

    # Default is Percentile with slider at 0.4%
    assert colormapPopup._resetButtonGroup.checkedButton() == percentileButton
    mode = colormapPopup._getResetMode()
    assert mode == "0.4percentile"

    QTest.mouseClick(colormapPopup._autoscaleButton, QtCore.Qt.LeftButton)
    range_ = colormapPopup.getRange()
    assert range_ == (0.4, 99.6)

    minmaxButton.click()
    mode = colormapPopup._getResetMode()
    range_ = colormapPopup.getRange()
    assert mode == "minmax"
    assert range_ == (0, 100)
    assert not colormapPopup._percentileSlider.isVisible()

    # Move slider to ~1% and switch back to percentile
    colormapPopup._percentileSlider.setValue(
        colormapPopup._percentile_to_slider(1.0)
    )
    percentileButton.click()
    mode = colormapPopup._getResetMode()
    assert "percentile" in mode
    range_ = colormapPopup.getRange()
    assert np.allclose(range_, (1.0, 99.0), atol=0.1)
    assert colormapPopup._percentileSlider.isVisible()

    mean3stdButton.click()
    mode = colormapPopup._getResetMode()
    range_ = colormapPopup.getRange()
    assert mode == "mean3std"
    assert range_ == (0, 100)
    assert not colormapPopup._percentileSlider.isVisible()



# SPDX-License-Identifier: MIT

import os
import sys
import pytest

import numpy as np
from mock import MagicMock

from pyFAI import detectors
from pyFAI.detectors import Detector
from pyFAI.detectors.orientation import Orientation

from ...model.CalibrationModel import (
    NoPointsError,
    get_available_detectors,
    DetectorModes,
)
from ... import calibrants_path

unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, "../data")


def load_pilatus_1M(img_model):
    img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))


def load_small_image(img_model, shape=(10, 10)):
    img_model._img_data = np.ones(shape)


def load_pilatus_1M_with_calibration(calibration_model):
    calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))
    load_pilatus_1M(calibration_model.img_model)


def load_small_image_with_calibration(calibration_model, shape=(10, 10)):
    calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))
    load_small_image(calibration_model.img_model, shape)


def load_image_with_distortion(calibration_model):
    calibration_model.img_model.load(
        os.path.join(data_path, "distortion", "CeO2_calib.edf")
    )
    calibration_model.find_peaks_automatic(1025.1, 1226.8, 0)
    calibration_model.set_calibrant(os.path.join(calibrants_path, "CeO2.D"))
    calibration_model.start_values["dist"] = 300e-3
    calibration_model.detector.pixel1 = 50e-6
    calibration_model.detector.pixel2 = 50e-6
    calibration_model.start_values["wavelength"] = 0.1e-10
    calibration_model.calibrate()


def load_LaB6_40keV_with_calibration(calibration_model):
    calibration_model.img_model.load(os.path.join(data_path, "image_001.tif"))
    calibration_model.load(os.path.join(data_path, "LaB6_40keV_MarCCD.poni"))


def test_load_poni_file_without_orientation(calibration_model, tmp_path):
    calibration_model.load(os.path.join(data_path, "LaB6_40keV_MarCCD.poni"))

    poni_config = calibration_model.pattern_geometry.get_config()
    orientation = poni_config["detector_config"].pop("orientation")

    assert orientation == Orientation.BottomRight

    # Save to temporary path instead of data_path
    temp_poni_file = os.path.join(tmp_path, "LaB6_40keV_MarCCD_new_with_orientation.poni")
    calibration_model.save(temp_poni_file)


    # Check that the orientation is still BottomRight
    calibration_model.load(temp_poni_file)

    poni_config = calibration_model.pattern_geometry.get_config()
    orientation = poni_config["detector_config"].pop("orientation")

    assert orientation == Orientation.BottomRight

    # Check that the orientation is TopRight in the poni file
    from pyFAI.io.ponifile import PoniFile
    poni_file = PoniFile(temp_poni_file)
    poni_dict = poni_file.as_dict()
    orientation = poni_dict["detector_config"]["orientation"]

    assert orientation == Orientation.TopRight


def test_load_poni_file_with_orientation(calibration_model):
    poni_file = os.path.join(data_path, "LaB6_40keV_MarCCD_new_with_orientation.poni")
    calibration_model.load(poni_file)

    poni_config = calibration_model.pattern_geometry.get_config()
    orientation = poni_config["detector_config"].pop("orientation")

    assert orientation == Orientation.BottomRight


def test_integration_with_supersampling(calibration_model):
    load_small_image_with_calibration(calibration_model)
    x1, y1 = calibration_model.integrate_1d()

    calibration_model.set_supersampling(2)
    x2, y2 = calibration_model.integrate_1d()

    assert len(y2) >= len(y1)
    y1_2_interp = np.interp(x2, x1, y1)

    assert np.mean((y2 - y1_2_interp)) == pytest.approx(0, abs=1e-2)


def test_get_pixel_ind(calibration_model):
    load_small_image_with_calibration(calibration_model, shape=(30, 30))
    calibration_model.integrate_1d(60)

    tth_array = calibration_model.tth_array
    azi_array = calibration_model.azi_array

    for _ in range(10):
        ind1 = np.random.randint(1, 10)
        ind2 = np.random.randint(1, 10)
        # the 0, 0 case is not working with the get_pixel_ind function

        tth = tth_array[ind1, ind2]
        azi = azi_array[ind1, ind2]

        result_ind1, result_ind2 = calibration_model.get_pixel_ind(tth, azi)

        assert ind1 == pytest.approx(result_ind1, abs=1e-3)
        assert ind2 == pytest.approx(result_ind2, abs=1e-3)


def test_use_different_image_sizes_for_1d_integration(calibration_model, img_model):
    load_small_image_with_calibration(calibration_model, shape=(10, 10))
    calibration_model.integrate_1d()
    load_small_image(img_model, shape=(12, 12))
    calibration_model.integrate_1d()


def test_use_different_image_sizes_for_2d_integration(calibration_model):
    load_small_image_with_calibration(calibration_model, shape=(10, 10))
    calibration_model.integrate_2d()
    load_small_image_with_calibration(calibration_model, shape=(12, 12))
    calibration_model.integrate_2d()


def test_correct_solid_angle(calibration_model, img_model):
    load_small_image_with_calibration(calibration_model, shape=(10, 10))
    _, y1 = calibration_model.integrate_1d()
    calibration_model.correct_solid_angle = False
    _, y2 = calibration_model.integrate_1d()
    assert np.sum(y1) != np.sum(y2)


def test_distortion_correction(calibration_model, img_model):
    load_image_with_distortion(calibration_model)

    _, y1 = calibration_model.integrate_1d()

    calibration_model.load_distortion(
        os.path.join(data_path, "distortion", "f4mnew.spline")
    )
    calibration_model.calibrate()

    _, y2 = calibration_model.integrate_1d()
    assert np.sum(y1) != np.sum(y2)
    assert y1[100] != pytest.approx(y2[100])


def test_cake_integral(calibration_model):
    load_pilatus_1M_with_calibration(calibration_model)
    calibration_model.integrate_2d(azimuth_points=360)

    cake_tth = calibration_model.cake_tth
    cake_img = calibration_model.cake_img
    cake_step = cake_tth[31] - cake_tth[30]

    # directly selecting value in the tth array
    _, y1 = calibration_model.cake_integral(cake_tth[30])
    assert np.array_equal(y1, calibration_model.cake_img[:, 30])

    # selecting exactly in between two points
    cake_partial = 0.5 * cake_img[:, 30] + 0.5 * cake_img[:, 31]
    _, y2 = calibration_model.cake_integral(cake_tth[30] + 0.5 * cake_step)
    assert np.allclose(y2, cake_partial, rtol=1e-6, atol=1e-6)

    # selecting points somewhere in between
    cake_partial = 0.3 * cake_img[:, 30] + 0.7 * cake_img[:, 31]
    _, y3 = calibration_model.cake_integral(cake_tth[30] + 0.7 * cake_step)
    assert np.allclose(y3, cake_partial, rtol=1e-6, atol=1e-6)

    # test with larger binsize of 2
    cake_partial = 0.5 * cake_img[:, 30] + 0.5 * cake_img[:, 31]
    _, y4 = calibration_model.cake_integral(cake_tth[30] + 0.5 * cake_step, bins=2)
    assert np.allclose(y4, cake_partial, rtol=1e-6, atol=1e-6)

    cake_partial = (0.5 * cake_img[:, 29] + cake_img[:, 30] + 0.5 * cake_img[:, 31]) / 2
    _, y5 = calibration_model.cake_integral(cake_tth[30], bins=2)
    assert np.allclose(y5, cake_partial, rtol=1e-6, atol=1e-6)


def test_integration_with_predefined_detector(calibration_model, img_model):
    calibration_model.load_detector("Pilatus CdTe 1M")
    calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))
    img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))
    assert len(calibration_model.tth) > 0


def test_integration_with_rotated_predefined_detector(calibration_model, img_model):
    load_pilatus_1M_with_calibration(calibration_model)
    calibration_model.load_detector("Pilatus CdTe 1M")
    x1, y1 = calibration_model.integrate_1d()

    # rotate m90
    calibration_model.rotate_detector_m90()
    img_model.rotate_img_m90()
    calibration_model.integrate_1d()

    # rotate p90
    calibration_model.rotate_detector_p90()
    img_model.rotate_img_p90()
    x2, y2 = calibration_model.integrate_1d()

    assert len(x1) == len(x2)
    assert float(np.sum((y1 - y2) ** 2)) == pytest.approx(0)


def test_integration_with_rotation(calibration_model, img_model):
    load_small_image_with_calibration(calibration_model, shape=(10, 12))
    calibration_model.integrate_1d()

    # rotate m90
    calibration_model.rotate_detector_m90()
    img_model.rotate_img_m90()
    calibration_model.integrate_1d()


def test_integration_with_rotation_and_reset(calibration_model, img_model):
    load_small_image_with_calibration(calibration_model, shape=(10, 12))
    calibration_model.load_detector("Pilatus CdTe 1M")
    x1, y1 = calibration_model.integrate_1d()

    calibration_model.rotate_detector_m90()
    img_model.rotate_img_m90()
    calibration_model.rotate_detector_m90()
    img_model.rotate_img_m90()
    calibration_model.rotate_detector_p90()
    img_model.rotate_img_p90()

    calibration_model.reset_transformations()
    img_model.reset_transformations()

    x2, y2 = calibration_model.integrate_1d()

    assert len(x1) == len(x2)
    assert float(np.sum((y1 - y2) ** 2)) == pytest.approx(0)

    calibration_model.rotate_detector_p90()
    img_model.rotate_img_p90()
    calibration_model.integrate_1d()


def test_integration_with_transformation_and_change_detector_to_custom(
    calibration_model, img_model
):
    load_small_image_with_calibration(calibration_model, shape=(10, 12))
    calibration_model.load_detector("Pilatus CdTe 1M")
    _ = calibration_model.integrate_1d()

    calibration_model.rotate_detector_m90()
    img_model.rotate_img_m90()


def test_change_detector_after_loading_image_with_different_shapes_integrate_1d(
    calibration_model, img_model
):
    load_small_image(img_model, shape=(10, 12))
    calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))
    calibration_model.integrate_1d()

    callback_function = MagicMock()
    calibration_model.detector_reset.connect(callback_function)
    calibration_model.load_detector("Pilatus CdTe 1M")
    calibration_model.integrate_1d()
    callback_function.assert_called_once()


def test_change_detector_after_loading_image_with_different_shapes_integrate_2d(
    calibration_model, img_model
):
    img_model._img_data = np.ones((10, 13))
    calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))
    calibration_model.integrate_1d()

    callback_function = MagicMock()
    calibration_model.detector_reset.connect(callback_function)
    calibration_model.load_detector("Pilatus CdTe 1M")
    calibration_model.integrate_2d()
    callback_function.assert_called_once()


def test_loading_calibration_gives_right_pixel_size(calibration_model):
    calibration_model.pattern_geometry.load(
        os.path.join(data_path, "CeO2_Pilatus1M.poni")
    )
    assert calibration_model.pattern_geometry.pixel1 == 0.000172

    calibration_model.load(os.path.join(data_path, "LaB6_40keV_MarCCD.poni"))
    assert calibration_model.pattern_geometry.pixel1 == 0.000079


def test_find_peaks_automatic(calibration_model, img_model):
    load_pilatus_1M(img_model)
    find_pilatus_1M_peaks(calibration_model)
    assert len(calibration_model.points) == 6
    for points in calibration_model.points:
        assert len(points) > 0


def test_find_peak(calibration_model, img_model):
    """
    Tests the find_peak function for several maxima and pick points

    """
    points_and_pick_points = [
        [[30, 50], [31, 49]],
        [[30, 50], [34, 46]],
        [[5, 5], [3, 3]],
        [[298, 298], [299, 299]],
    ]

    for data in points_and_pick_points:
        img_model._img_data = np.zeros((300, 300))

        point = data[0]
        pick_point = data[1]
        img_model._img_data[point[0], point[1]] = 100

        peak_point = calibration_model.find_peak(pick_point[0], pick_point[1], 10, 0)
        assert peak_point[0][0] == point[0]
        assert peak_point[0][1] == point[1]


def find_pilatus_1M_peaks(calibration_model):
    points = [
        (517.664434674, 646, 0),
        (667.380513299, 525.252854758, 0),
        (671.110095329, 473.571503774, 0),
        (592.788872703, 350.495296791, 0),
        (387.395462348, 390.987901686, 0),
        (367.94835605, 554.290314848, 0),
    ]
    for point in points:
        calibration_model.find_peaks_automatic(point[0], point[1], 0)


def test_calibration_with_supersampling1(calibration_model, img_model):
    load_pilatus_1M(img_model)
    find_pilatus_1M_peaks(calibration_model)
    calibration_model.set_calibrant(os.path.join(calibrants_path, "CeO2.D"))
    calibration_model.detector.pixel1 = 172e-6
    calibration_model.detector.pixel2 = 172e-6

    calibration_model.calibrate()
    normal_poni1 = calibration_model.pattern_geometry.poni1
    normal_poni2 = calibration_model.pattern_geometry.poni2

    calibration_model.set_supersampling(2)

    calibration_model.calibrate()
    assert pytest.approx(normal_poni1) == calibration_model.pattern_geometry.poni1
    assert pytest.approx(normal_poni2) == calibration_model.pattern_geometry.poni2


def test_calibration_with_supersampling2(calibration_model, img_model):
    load_pilatus_1M(img_model)
    calibration_model.set_calibrant(os.path.join(calibrants_path, "CeO2.D"))
    calibration_model.detector.pixel1 = 172e-6
    calibration_model.detector.pixel2 = 172e-6

    calibration_model.set_supersampling(2)
    find_pilatus_1M_peaks(calibration_model)

    calibration_model.calibrate()
    super_poni1 = calibration_model.pattern_geometry.poni1
    super_poni2 = calibration_model.pattern_geometry.poni2

    calibration_model.set_supersampling(1)
    find_pilatus_1M_peaks(calibration_model)

    calibration_model.calibrate()
    assert (
        pytest.approx(super_poni1, abs=1e-3) == calibration_model.pattern_geometry.poni1
    )
    assert (
        pytest.approx(super_poni2, abs=1e-3) == calibration_model.pattern_geometry.poni2
    )


def test_calibration1(calibration_model, img_model):
    img_model.load(os.path.join(data_path, "LaB6_40keV_MarCCD.tif"))
    calibration_model.find_peaks_automatic(1179.6, 1129.4, 0)
    calibration_model.find_peaks_automatic(1268.5, 1119.8, 1)

    calibration_model.set_calibrant(os.path.join(calibrants_path, "LaB6.D"))
    calibration_model.calibrate()

    assert calibration_model.pattern_geometry.poni1 > 0
    assert calibration_model.pattern_geometry.dist == pytest.approx(0.18, abs=0.01)
    assert calibration_model.cake_geometry.poni1 > 0


def test_calibration2(calibration_model, img_model):
    img_model.load(os.path.join(data_path, "LaB6_OffCenter_PE.tif"))
    calibration_model.find_peaks_automatic(1245.2, 1919.3, 0)
    calibration_model.find_peaks_automatic(1334.0, 1823.7, 1)
    calibration_model.set_start_values(
        {"dist": 500e-3, "polarization_factor": 0.99, "wavelength": 0.3344e-10}
    )
    calibration_model.set_pixel_size((200e-6, 200e-6))
    calibration_model.set_calibrant(os.path.join(calibrants_path, "LaB6.D"))
    calibration_model.calibrate()

    assert calibration_model.pattern_geometry.poni1 > 0
    assert calibration_model.pattern_geometry.dist == pytest.approx(0.500, abs=0.01)
    assert calibration_model.cake_geometry.poni1 > 0


def test_calibration3(calibration_model, img_model):
    load_pilatus_1M(img_model)
    find_pilatus_1M_peaks(calibration_model)
    calibration_model.start_values["wavelength"] = 0.406626e-10
    calibration_model.set_start_values(
        {"dist": 200e-3, "polarization_factor": 0.99, "wavelength": 0.406626e-10}
    )
    calibration_model.set_pixel_size((172e-6, 172e-6))
    calibration_model.set_calibrant(os.path.join(calibrants_path, "CeO2.D"))
    calibration_model.calibrate()

    assert calibration_model.pattern_geometry.poni1 > 0
    assert calibration_model.pattern_geometry.dist == pytest.approx(0.208, abs=0.005)
    assert calibration_model.cake_geometry.poni1 > 0


def test_calibration_with_fixed_parameters(calibration_model, img_model):
    load_pilatus_1M(img_model)
    find_pilatus_1M_peaks(calibration_model)
    calibration_model.start_values["wavelength"] = 0.406626e-10
    calibration_model.detector.pixel1 = 172e-6
    calibration_model.detector.pixel2 = 172e-6
    calibration_model.set_calibrant(os.path.join(calibrants_path, "CeO2.D"))

    fixed_values_dicts = [
        {"rot1": 0.001},
        {"rot2": 0.03},
        {"rot1": 0.01, "rot2": 0.003},
        {"poni1": 0.32},
        {"poni1": 0.2, "poni2": 0.13},
        {"dist": 300},
        {"rot1": 0.001, "rot2": 0.004, "poni1": 0.22, "poni2": 0.34},
    ]
    for fixed_values in fixed_values_dicts:
        calibration_model.set_fixed_values(fixed_values)
        calibration_model.calibrate()
        for key, value in fixed_values.items():
            assert getattr(calibration_model.pattern_geometry, key) == value


def test_get_two_theta_img_with_distortion(calibration_model):
    load_image_with_distortion(calibration_model)

    x, y = np.array((100,)), np.array((100,))
    calibration_model.get_two_theta_img(x, y)
    calibration_model.load_distortion(
        os.path.join(data_path, "distortion", "f4mnew.spline")
    )
    calibration_model.get_two_theta_img(x, y)


def test_cake_integration_with_small_azimuth_range(calibration_model, img_model):
    img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))
    calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))

    full_cake = calibration_model.integrate_2d()
    small_cake = calibration_model.integrate_2d(azimuth_range=(40, 130))
    assert not np.array_equal(full_cake, small_cake)


def test_cake_integration_with_off_azimuth_range(calibration_model, img_model):
    load_pilatus_1M_with_calibration(calibration_model)
    calibration_model.integrate_2d(azimuth_range=(150, -130))

    assert np.min(calibration_model.cake_azi) > 150
    assert np.max(calibration_model.cake_azi) < 230


def test_cake_integration_with_different_num_points(calibration_model, img_model):
    load_pilatus_1M_with_calibration(calibration_model)

    calibration_model.integrate_2d(rad_points=200)
    assert len(calibration_model.cake_tth) == 200

    calibration_model.integrate_2d(azimuth_points=200)
    assert len(calibration_model.cake_azi) == 200


def test_transforms_without_predefined_detector(calibration_model, img_model):
    img_model.load(os.path.join(data_path, "image_001.tif"))
    calibration_model.rotate_detector_p90()
    calibration_model.rotate_detector_m90()
    calibration_model.flip_detector_horizontally()
    calibration_model.img_model.flip_img_horizontally()
    calibration_model.flip_detector_horizontally()


def test_transforms_without_predefined_detector_changing_shape(
    calibration_model, img_model
):
    img_model.load(os.path.join(data_path, "image_001.tif"))
    calibration_model.rotate_detector_p90()
    img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))
    calibration_model.rotate_detector_m90()
    calibration_model.flip_detector_horizontally()
    calibration_model.img_model.flip_img_horizontally()
    calibration_model.flip_detector_horizontally()


def test_load_detector_list():
    names, classes = get_available_detectors()
    for name, cls in zip(names, classes):
        if name.startswith("Quantum"):
            assert "ADSC_" in str(cls)
        elif name.startswith("aca1300"):
            assert "Basler" in str(cls)
        else:
            assert name[:2].lower() in str(cls).lower()

    assert "Detector" not in names


def test_load_predefined_detector(calibration_model):
    calibration_model.load_detector("MAR 345")

    assert calibration_model.orig_pixel1 == 100e-6
    assert calibration_model.detector.pixel1 == 100e-6


def test_load_predefined_detector_and_poni_after(calibration_model):
    calibration_model.load_detector("Pilatus CdTe 1M")
    assert isinstance(calibration_model.detector, detectors.PilatusCdTe1M)
    assert isinstance(
        calibration_model.pattern_geometry.detector, detectors.PilatusCdTe1M
    )

    calibration_model.load(os.path.join(data_path, "CeO2_Pilatus1M.poni"))
    assert isinstance(calibration_model.detector, detectors.PilatusCdTe1M)
    assert isinstance(
        calibration_model.pattern_geometry.detector, detectors.PilatusCdTe1M
    )


def test_load_predefined_detector_and_poni_with_different_pixel_size(calibration_model):
    calibration_model.load_detector("Pilatus CdTe 1M")
    assert isinstance(calibration_model.detector, detectors.PilatusCdTe1M)
    assert isinstance(
        calibration_model.pattern_geometry.detector, detectors.PilatusCdTe1M
    )

    calibration_model.load(os.path.join(data_path, "LaB6_40keV_MarCCD.poni"))
    assert calibration_model.detector_mode == DetectorModes.CUSTOM
    assert isinstance(calibration_model.detector, detectors.Detector)
    assert isinstance(calibration_model.pattern_geometry.detector, detectors.Detector)


def test_load_detector_from_file(calibration_model):
    calibration_model.load_detector_from_file(os.path.join(data_path, "detector.h5"))
    assert calibration_model.orig_pixel1 == pytest.approx(100e-6)
    assert calibration_model.orig_pixel2 == pytest.approx(100e-6)
    assert calibration_model.detector.pixel1 == pytest.approx(100e-6)
    assert calibration_model.detector.pixel2 == pytest.approx(100e-6)
    assert calibration_model.detector.shape == (1048, 1032)


def test_load_detector_with_spline_file(calibration_model, tmp_path):
    # create detector and save it
    spline_detector = Detector()
    spline_detector.splinefile = os.path.join(
        data_path, "distortion", "f4mnew.spline"
    )
    spline_detector.save(os.path.join(tmp_path, "detector_with_spline.h5"))

    # load and check if it is working
    calibration_model.load_detector_from_file(
        os.path.join(tmp_path, "detector_with_spline.h5")
    )
    detector = calibration_model.detector
    assert detector.pixel1 == pytest.approx(50e-6)
    assert not detector.uniform_pixel


def test_calibrate_without_points(calibration_model):
    with pytest.raises(NoPointsError):
        calibration_model.calibrate()


def test_refine_without_points(calibration_model):
    with pytest.raises(NoPointsError):
        calibration_model.refine()


def test_dioptrin_integrator_recreated_on_image_shape_change(calibration_model, img_model):
    """Switching to an image with a different shape must recreate the dioptrin integrator."""
    load_pilatus_1M_with_calibration(calibration_model)

    calibration_model.use_dioptrin = True
    calibration_model._dioptrin_integrator = MagicMock()
    calibration_model._create_dioptrin_integrator = MagicMock()

    # Loading an image with a different shape should trigger recreation
    img_model.load(os.path.join(data_path, "image_001.tif"))
    calibration_model._create_dioptrin_integrator.assert_called_once()

    # Loading an image with the same shape should NOT trigger recreation
    calibration_model._create_dioptrin_integrator.reset_mock()
    img_model._img_data = np.ones(img_model.img_data.shape)
    img_model.img_changed.emit()
    calibration_model._create_dioptrin_integrator.assert_not_called()



# SPDX-License-Identifier: MIT

import inspect
import weakref

__export__ = ["Signal"]


class Signal:
    def __init__(self, *_):
        self.listeners = WeakRefList()
        self.priority_listeners = WeakRefList()
        self.blocked = False

    def connect(self, handle, priority=False):
        """
        Connects a function handle to the Signal.
        :param handle: function handle to be called when the signal is emitted
        :param priority: if set to True, the handle will be added as the first function to be called in the case of
                        the event
        If multiple handles are added with priority, they will obviously be called in the reverse order of adding.
        """
        if priority:
            self.priority_listeners.insert(0, handle)
        else:
            self.listeners.append(handle)

    def disconnect(self, handle):
        """
        Removes a certain function handle from the list of listeners to be called in case of Signal emitted.
        :param handle: function handle to be removed from the listeners
        """
        try:
            self.listeners.remove(handle)
        except ValueError:
            pass

        try:
            self.priority_listeners.remove(handle)
        except ValueError:
            pass

    def emit(self, *args):
        if self.blocked:
            return
        self._serve_listeners(self.priority_listeners, *args)
        self._serve_listeners(self.listeners, *args)

    @staticmethod
    def _serve_listeners(listeners, *args):
        for ref in listeners:
            handle = ref()
            if type(handle) == Signal:
                handle.emit(*args)
            else:
                if len(inspect.signature(handle).parameters) == 0:
                    handle()
                else:
                    handle(*args)

    def clear(self):
        """
        Removes all listeners from the Signal.
        """
        self.listeners = WeakRefList()
        self.priority_listeners = WeakRefList()

    def block(self):
        """
        Blocks the Signal from emitting.
        """
        self.blocked = True

    def unblock(self):
        """
        Unblocks the Signal from emitting.
        """
        self.blocked = False

    def has_listener(self, handle):
        """
        Returns True if the handle is in the list of listeners.
        """
        return handle in self.listeners or handle in self.priority_listeners


class WeakRefList(list):
    """
    A list which holds weak references to its items. If an item is deleted, the reference to it will be removed from
    the list. This is useful for Signals, where we want to hold a list of listeners, but don't want to prevent the
    garbage collector from deleting the listeners.
    It is not a full reimplementation, only the methods which are used in the Signal class are implemented - append,
    remove, insert. This list will work for object methods as well as objects. To retrieve the orginal item, the
    value of the weak reference has to be called. E.g.:
    >>> class A:
    >>>     def method(self):
    >>>         return "lala"
    >>>
    >>> a = A()
    >>> weak_ref_list = WeakRefList()
    >>> weak_ref_list.append(a.method)
    >>> weak_ref_list[0]()() == "lala"
    """

    def append(self, item):
        super(WeakRefList, self).append(self._ref(item))

    def remove(self, item):
        super(WeakRefList, self).remove(self._ref(item))

    def insert(self, index, item):
        super(WeakRefList, self).insert(index, self._ref(item))

    def _remove_ref(self, ref):
        super(WeakRefList, self).remove(ref)

    def _ref(self, item):
        if inspect.ismethod(item):
            return weakref.WeakMethod(item, self._remove_ref)
        else:
            return weakref.ref(item, self._remove_ref)

    def __contains__(self, item):
        for ref in self:
            if ref() == item:
                return True
        return False



# SPDX-License-Identifier: MIT

import pytest
from mock import MagicMock, patch
import os

import numpy as np

from ...model.ImgModel import ImgModel, BackgroundDimensionWrongException
from ...model.util.ImgCorrection import DummyCorrection
from ...model.loader.KaraboLoader import extra_data_installed

unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, "../data")
spe_path = os.path.join(data_path, "spe")


@pytest.fixture
def img_model():
    img_model = ImgModel()
    img_model.load(os.path.join(data_path, "image_001.tif"))
    return ImgModel()


def test_load_karabo_nexus_file(img_model):
    img_model.load(os.path.join(data_path, "karabo_epix.h5"))


def perform_transformations_tests(img_model):
    assert np.sum(np.absolute(img_model.img_data)) == 0
    img_model.rotate_img_m90()
    assert np.sum(np.absolute(img_model.img_data)) == 0
    img_model.flip_img_horizontally()
    assert np.sum(np.absolute(img_model.img_data)) == 0
    img_model.rotate_img_p90()
    assert np.sum(np.absolute(img_model.img_data)) == 0
    img_model.flip_img_vertically()
    assert np.sum(np.absolute(img_model.img_data)) == 0
    img_model.reset_transformations()
    assert np.sum(np.absolute(img_model.img_data)) == 0


def test_load_emits_signal(img_model):
    callback_fcn = MagicMock()
    img_model.img_changed.connect(callback_fcn)
    img_model.load(os.path.join(data_path, "image_001.tif"))
    callback_fcn.assert_called_once_with()


def test_flipping_images(img_model):
    original_image = np.copy(img_model._img_data)
    img_model.flip_img_vertically()
    assert np.array_equal(img_model._img_data, np.flipud(original_image))


def test_simple_background_subtraction(img_model):
    first_image = np.copy(img_model.img_data)
    img_model.load_next_file()
    second_image = np.copy(img_model.img_data)

    img_model.load(os.path.join(data_path, "image_001.tif"))
    img_model.load_background(os.path.join(data_path, "image_002.tif"))

    assert not np.array_equal(first_image, img_model.img_data)

    img_model.load_next_file()
    assert np.sum(img_model.img_data) == 0


def test_background_subtraction_with_transformation(img_model):
    img_model.load_background(os.path.join(data_path, "image_002.tif"))
    original_img = np.copy(img_model._img_data)
    original_background = np.copy(img_model._background_data)

    assert img_model._background_data is not None
    assert not np.array_equal(img_model.img_data, img_model._img_data)

    original_img_background_subtracted = np.copy(img_model.img_data)
    assert np.array_equal(
        original_img_background_subtracted, original_img - original_background
    )

    ### now comes the main process - flipping the image
    img_model.flip_img_vertically()
    flipped_img = np.copy(img_model._img_data)
    assert np.array_equal(np.flipud(original_img), flipped_img)

    flipped_background = np.copy(img_model._background_data)
    assert np.array_equal(np.flipud(original_background), flipped_background)

    flipped_img_background_subtracted = np.copy(img_model.img_data)
    assert np.array_equal(
        flipped_img_background_subtracted, flipped_img - flipped_background
    )

    assert np.array_equal(
        np.flipud(original_img_background_subtracted), flipped_img_background_subtracted
    )
    assert (
        np.sum(
            np.flipud(original_img_background_subtracted)
            - flipped_img_background_subtracted
        )
        == 0
    )

    img_model.load(os.path.join(data_path, "image_002.tif"))
    perform_transformations_tests(img_model)


def test_background_scaling_and_offset(img_model):
    img_model.load_background(os.path.join(data_path, "image_002.tif"))

    # assure that everything is correct before
    assert np.array_equal(
        img_model.img_data, img_model._img_data - img_model._background_data
    )

    # set scaling and see difference
    img_model.background_scaling = 2.4
    assert np.array_equal(
        img_model.img_data, img_model._img_data - 2.4 * img_model._background_data
    )

    # set offset and see the difference
    img_model.background_scaling = 1.0
    img_model.background_offset = 100.0
    assert np.array_equal(
        img_model.img_data, img_model._img_data - (img_model._background_data + 100.0)
    )

    # use offset and scaling combined
    img_model.background_scaling = 2.3
    img_model.background_offset = 100.0
    assert np.array_equal(
        img_model.img_data,
        img_model._img_data - (2.3 * img_model._background_data + 100),
    )


def test_background_with_different_shape(img_model):
    with pytest.raises(BackgroundDimensionWrongException):
        img_model.load_background(os.path.join(data_path, "CeO2_Pilatus1M.tif"))
    assert img_model._background_data is None

    img_model.load_background(os.path.join(data_path, "image_002.tif"))
    assert img_model._background_data is not None

    img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))
    assert img_model._background_data is None


def test_absorption_correction_with_different_image_sizes(img_model):
    dummy_correction = DummyCorrection(img_model.img_data.shape, 0.4)
    # self.img_data.set_absorption_correction(np.ones(self.img_data._img_data.shape)*0.4)
    img_model.add_img_correction(dummy_correction, "Dummy 1")
    assert img_model._img_corrections.has_items()

    img_model.load(os.path.join(data_path, "CeO2_Pilatus1M.tif"))
    assert not img_model.has_corrections()


def test_adding_several_absorption_corrections(img_model):
    original_image = np.copy(img_model.img_data)
    img_shape = original_image.shape
    img_model.add_img_correction(DummyCorrection(img_shape, 0.4))
    img_model.add_img_correction(DummyCorrection(img_shape, 3))
    img_model.add_img_correction(DummyCorrection(img_shape, 5))

    assert np.sum(original_image) / (0.5 * 3 * 5) == np.sum(img_model.img_data)

    img_model.delete_img_correction(1)
    assert np.sum(original_image) / (0.5 * 5) == np.sum(img_model.img_data)


def test_saving_data(img_model, tmp_path):
    img_model.load(os.path.join(data_path, "image_001.tif"))
    filename = os.path.join(tmp_path, "test.tif")
    img_model.save(filename)
    first_img_array = np.copy(img_model._img_data)
    img_model.load(filename)
    assert np.array_equal(first_img_array, img_model._img_data)
    assert os.path.exists(filename)


def test_negative_rotation(img_model):
    pre_transformed_data = img_model.img_data
    img_model.rotate_img_m90()
    img_model.rotate_img_m90()
    img_model.rotate_img_m90()
    img_model.rotate_img_m90()
    assert np.array_equal(img_model.img_data, pre_transformed_data)


def test_combined_rotation(img_model):
    pre_transformed_data = img_model.img_data
    img_model.rotate_img_m90()
    img_model.rotate_img_p90()
    assert np.array_equal(img_model.img_data, pre_transformed_data)


def test_flip_img_horizontally(img_model):
    pre_transformed_data = img_model.img_data
    img_model.flip_img_horizontally()
    img_model.flip_img_horizontally()
    assert np.array_equal(img_model.img_data, pre_transformed_data)


def test_flip_img_vertically(img_model):
    pre_transformed_data = img_model.img_data
    img_model.flip_img_vertically()
    img_model.flip_img_vertically()
    assert np.array_equal(img_model.img_data, pre_transformed_data)


def test_combined_rotation_and_flipping(img_model):
    img_model.load(os.path.join(data_path, "image_001.tif"))
    img_model.flip_img_vertically()
    img_model.flip_img_horizontally()
    img_model.rotate_img_m90()
    img_model.rotate_img_p90()
    img_model.rotate_img_m90()
    img_model.rotate_img_m90()
    img_model.flip_img_horizontally()
    transformed_data = img_model.img_data
    img_model.load(os.path.join(data_path, "image_001.tif"))
    assert np.array_equal(img_model.img_data, transformed_data)


def test_reset_img_transformation(img_model):
    pre_transformed_data = img_model.img_data
    img_model.rotate_img_m90()
    img_model.reset_transformations()
    assert np.array_equal(img_model.img_data, pre_transformed_data)

    pre_transformed_data = img_model.img_data
    img_model.rotate_img_p90()
    img_model.reset_transformations()
    assert np.array_equal(img_model.img_data, pre_transformed_data)

    pre_transformed_data = img_model.img_data
    img_model.flip_img_horizontally()
    img_model.reset_transformations()
    assert np.array_equal(img_model.img_data, pre_transformed_data)

    pre_transformed_data = img_model.img_data
    img_model.flip_img_vertically()
    img_model.reset_transformations()
    assert np.array_equal(img_model.img_data, pre_transformed_data)

    pre_transformed_data = img_model.img_data
    img_model.flip_img_vertically()
    img_model.flip_img_horizontally()
    img_model.rotate_img_m90()
    img_model.rotate_img_p90()
    img_model.rotate_img_m90()
    img_model.rotate_img_m90()
    img_model.flip_img_horizontally()
    img_model.reset_transformations()
    assert np.array_equal(img_model.img_data, pre_transformed_data)


def test_loading_a_tagged_tif_file_and_retrieving_info_string(img_model):
    img_model.load(os.path.join(data_path, "attrib.tif"))
    assert "areaDetector" in img_model.file_info


def test_loading_spe_file(img_model):
    img_model.load(os.path.join(spe_path, "CeO2_PI_CCD_Mo.SPE"))
    assert img_model.img_data.shape == (1042, 1042)


def test_loading_ESRF_hdf5_file(img_model):
    img_model.load(os.path.join(data_path, "hdf5_dataset", "ma4500_demoh5.h5"))
    assert img_model.img_data.shape == (2048, 2048)

    img1 = img_model.img_data
    img_model.select_source(img_model.sources[2])
    img2 = img_model.img_data
    assert np.sum(img1 - img2) != 0


def test_summing_files(img_model):
    img_model.load(os.path.join(data_path, "image_001.tif"))
    data1 = np.copy(img_model._img_data).astype(np.uint64)
    img_model.add(os.path.join(data_path, "image_001.tif"))
    assert np.array_equal(2 * data1, img_model._img_data)


def test_summing_rotated(img_model):
    img_model.load(os.path.join(data_path, "image_001.tif"))
    img_model.rotate_img_m90()
    data1 = np.copy(img_model._img_data).astype(np.uint32)
    img_model.add(os.path.join(data_path, "image_001.tif"))
    assert np.array_equal(2 * data1, img_model._img_data)


def test_loading_karabo_file(img_model):
    img_model.load(os.path.join(data_path, "karabo_epix.h5"))
    assert img_model.img_data.shape == (356, 384)


def test_loading_karabo_file_without_extra_data(img_model):
    """Test that loading a karabo file when extra_data is not installed returns None"""
    with patch("dioptas.model.loader.KaraboLoader.extra_data_installed", False):
        # Try to load a karabo file
        result = img_model.load(os.path.join(data_path, "karabo_epix.h5"))
        assert result is None

        # Verify that other file types still work
        img_model.load(os.path.join(data_path, "image_001.tif"))
        assert img_model.img_data is not None



# SPDX-License-Identifier: MIT
import os
import time
from PIL import Image
from typing import Optional

import numpy as np
import pyqtgraph as pg
from qtpy import QtWidgets, QtCore
from qtpy.QtGui import QTransform
from scipy.ndimage import zoom

from dioptas.model.DioptasModel import DioptasModel
from dioptas.model.util.calc import convert_units
from dioptas.widgets.MapWidget import MapWidget

from ..widgets.UtilityWidgets import get_progress_dialog, open_files_dialog
from .integration.phase.PhaseInPatternController import PhaseInPatternController
from .integration.overlay.OverlayInPatternController import OverlayInPatternController
from ..widgets.UtilityWidgets import save_file_dialog


class MapController(object):
    def __init__(self, widget: MapWidget, dioptas_model: DioptasModel):
        self.widget = widget
        self.model = dioptas_model

        self._contour_items: list[pg.IsocurveItem] = []
        self._setting_levels = False

        self.phase_in_pattern_controller = PhaseInPatternController(
            self.widget.pattern_plot_widget, self.model
        )
        self.overlay_in_pattern_controller = OverlayInPatternController(
            self.widget.pattern_plot_widget, self.model.overlay_model
        )

        self.create_signals()

    def create_signals(self):
        self.widget.control_widget.load_btn.clicked.connect(self.load_btn_clicked)
        self.widget.control_widget.file_list.currentRowChanged.connect(
            self.file_list_row_changed
            # needs to be its own function, to always recall the model.map_model
            # this ensures, that the currently selected configuration is used
        )
        self.widget.control_widget.reintegrate_cb.toggled.connect(
            self._auto_integrate_toggled
        )
        self.widget.map_plot_control_widget.save_map_btn.clicked.connect(self._save_map)
        self.widget.map_image_frame.smooth_btn.toggled.connect(
            self._smooth_toggled
        )
        self.widget.map_image_frame.smooth_slider.valueChanged.connect(
            self._smooth_slider_changed
        )
        self.widget.map_image_frame.contour_btn.toggled.connect(
            self._contour_toggled
        )
        self.widget.map_image_frame.contour_slider.valueChanged.connect(
            self._contour_slider_changed
        )

        self.widget.pattern_footer_widget.log_btn.clicked.connect(
            self._y_scale_log_clicked
        )
        self.widget.pattern_footer_widget.sqrt_btn.clicked.connect(
            self._y_scale_sqrt_clicked
        )

        self.widget.map_plot_widget.mouse_left_clicked.connect(self.map_point_selected)
        self.widget.pattern_plot_widget.mouse_left_clicked.connect(self.pattern_clicked)
        self.widget.pattern_plot_widget.map_interactive_roi.sigRegionChanged.connect(
            self.pattern_roi_changed
        )
        self.widget.pattern_plot_widget.mouse_moved.connect(
            self.pattern_plot_mouse_moved
        )

        self.widget.map_plot_control_widget.map_dimension_cb.currentIndexChanged.connect(
            self.map_dimension_cb_changed
        )
        self.widget.map_plot_widget.mouse_moved.connect(self.map_plot_mouse_moved)
        self.widget.img_plot_widget.mouse_left_clicked.connect(
            self.img_plot_left_clicked
        )
        self.widget.img_plot_widget.mouse_moved.connect(self.img_plot_mouse_moved)
        self.widget.img_autoscale_btn.clicked.connect(
            self._img_autoscale_btn_clicked
        )
        self.widget.img_plot_widget.img_histogram_LUT_horizontal.sigLevelChangeFinished.connect(
            self._img_levels_manually_changed
        )

        self.model.map_model.map_changed.connect(self.update_map)
        self.model.map_model.map_changed.connect(self.update_file_list)
        self.model.clicked_tth_changed.connect(self.update_pattern_green_line)
        self.model.clicked_tth_changed.connect(self.update_image_green_line)
        self.model.clicked_tth_changed.connect(self.update_clicked_pos_label)
        self.model.clicked_azi_changed.connect(self.update_clicked_azi_label)

        self.model.pattern_changed.connect(self.update_pattern)
        self.activate_model_signals()

    def _smooth_toggled(self, checked: bool):
        self.widget.map_image_frame.smooth_slider.setVisible(checked)
        self.widget.map_image_frame.smooth_label.setVisible(checked)
        if checked:
            factor = self.widget.map_image_frame.smooth_slider.value()
        else:
            factor = 1
        self.widget.map_plot_widget.data_img_item.setSmoothFactor(factor)

    def _smooth_slider_changed(self, value: int):
        self.widget.map_image_frame.smooth_label.setText(str(value))
        if self.widget.map_image_frame.smooth_btn.isChecked():
            self.widget.map_plot_widget.data_img_item.setSmoothFactor(value)

    def _contour_slider_changed(self, value: int):
        self.widget.map_image_frame.contour_label.setText(str(value))
        if self.widget.map_image_frame.contour_btn.isChecked():
            self._update_contours()

    def _contour_toggled(self, checked: bool):
        self.widget.map_image_frame.contour_slider.setVisible(checked)
        self.widget.map_image_frame.contour_label.setVisible(checked)
        if checked:
            self._update_contours()
        else:
            self._clear_contours()

    _CONTOUR_UPSAMPLE = 3

    def _update_contours(self):
        self._clear_contours()
        if self.model.map_model.map is None:
            return
        data = np.flipud(self.model.map_model.map).T
        factor = self._CONTOUR_UPSAMPLE
        data = zoom(data.astype(float), factor, order=3)
        num_levels = self.widget.map_image_frame.contour_slider.value()
        d_min, d_max = float(data.min()), float(data.max())
        if d_min == d_max:
            return
        levels = np.linspace(d_min, d_max, num_levels + 2)[1:-1]
        pen = pg.mkPen(color=(255, 255, 255, 128), width=1)
        view_box = self.widget.map_plot_widget.img_view_box
        scale = QTransform.fromScale(1.0 / factor, 1.0 / factor)
        for level in levels:
            item = pg.IsocurveItem(data=data, level=level, pen=pen)
            item.setTransform(scale)
            view_box.addItem(item)
            self._contour_items.append(item)

    def _clear_contours(self):
        view_box = self.widget.map_plot_widget.img_view_box
        for item in self._contour_items:
            view_box.removeItem(item)
        self._contour_items.clear()

    def activate(self):
        self.activate_model_signals()
        self._apply_auto_integrate()
        self.configuration_selected()

    def activate_model_signals(self):
        self.model.img_changed.connect(self.update_image)
        self.model.configuration_selected.connect(self.configuration_selected)

    def deactivate(self):
        self.model.img_changed.disconnect(self.update_image)
        self.model.configuration_selected.disconnect(self.configuration_selected)
        self.model.current_configuration.auto_integrate_pattern = True

    def _auto_integrate_toggled(self, checked):
        self.model.current_configuration.auto_integrate_pattern = checked
        if checked:
            self.model.current_configuration.integrate_image_1d()
        else:
            ind = self.widget.control_widget.file_list.currentRow()
            if ind >= 0:
                self._set_stored_pattern(ind)

    def _img_autoscale_btn_clicked(self):
        if self.widget.img_autoscale_btn.isChecked():
            self._setting_levels = True
            self.widget.img_plot_widget.auto_level()
            self._setting_levels = False

    def _img_levels_manually_changed(self, *args):
        if not self._setting_levels:
            self.widget.img_autoscale_btn.setChecked(False)

    def _apply_auto_integrate(self):
        checked = self.widget.control_widget.reintegrate_cb.isChecked()
        self.model.current_configuration.auto_integrate_pattern = checked

    def _set_stored_pattern(self, index):
        """Set pattern from stored map data when reintegrate is off."""
        if self.widget.control_widget.reintegrate_cb.isChecked():
            return
        map_model = self.model.map_model
        if map_model.pattern_x is None or map_model.pattern_intensities is None:
            return
        x = map_model.pattern_x
        y = map_model.pattern_intensities[index]
        filename = self.model.img_model.filename
        self.model.current_configuration.pattern_model.set_pattern(
            x, y, filename, unit=self.model.integration_unit
        )

    def load_btn_clicked(self):
        filenames = open_files_dialog(
            self.widget,
            "Load image data file(s)",
            self.model.working_directories["image"],
        )
        if len(filenames) == 0:
            return

        progressDialog = get_progress_dialog(
            "Integrating image data...", "Abort Integration", 100, self.widget.map_pg_layout
        )
        progressDialog.setMinimumDuration(0)
        progressDialog.setWindowModality(QtCore.Qt.ApplicationModal)
        label = progressDialog.findChild(QtWidgets.QLabel)
        if label is not None:
            label.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)

        t_start = time.time()

        def callback_fn(current, n_total):
            if progressDialog.wasCanceled():
                return False
            progressDialog.setValue(int(current / n_total * 100))
            elapsed = time.time() - t_start
            rate = current / elapsed if elapsed > 0 else 0
            progressDialog.setLabelText(
                f"Image {current} of {n_total}\n"
                f"{elapsed:.1f}s elapsed\n"
                f"{rate:.1f} img/s"
            )
            QtWidgets.QApplication.processEvents()
            return not progressDialog.wasCanceled()

        try:
            self.model.map_model.load(filenames, callback_fn=callback_fn)
            self.model.map_model.select_point(0, 0)
        except Exception as e:
            QtWidgets.QMessageBox.critical(
                self.widget, "Error loading image data.", str(e)
            )
        finally:
            progressDialog.close()

    def _save_map(self):
        filename = save_file_dialog(
            self.widget,
            "Save Image.",
            os.path.join(self.model.working_directories["image"]),
            ("PNG Image (*.png);; TIFF Data (*.tiff);; Tabular Text (*.txt)"),
        )

        if filename == "":
            return
        else:
            self.save_map(filename)

    def save_map(self, filename: str):
        if filename.endswith(".png"):
            self.widget.map_plot_widget.save_img(filename)
        elif filename.endswith(".tiff"):
            data = self.model.map_model.map
            max_uint32 = np.iinfo(np.uint32).max
            normalized_data = (data - np.min(data)) / (np.max(data) - np.min(data))
            normalized_data = (normalized_data * max_uint32).astype(np.uint32)
            im = Image.fromarray(normalized_data)
            im.save(filename)
        elif filename.endswith(".txt"):
            np.savetxt(filename, self.model.map_model.map, fmt="%d")

    def update_file_list(self):
        # get current items
        items = [
            self.widget.control_widget.file_list.item(i).text()
            for i in range(self.widget.control_widget.file_list.count())
        ]
        if items == self.model.map_model.get_filenames():
            return

        self.widget.control_widget.file_list.blockSignals(True)
        self.widget.control_widget.file_list.clear()
        filenames = self.model.map_model.get_filenames()
        if len(filenames) == 0:  # no files loaded
            return
        self.widget.control_widget.file_list.addItems(filenames)
        self.widget.control_widget.file_list.blockSignals(False)

    def update_map(self):
        if self.model.map_model.map is None:
            # clear image
            self.widget.map_plot_widget.plot_image(np.array([[], []]))
            self._clear_contours()
        else:
            self.widget.map_plot_widget.plot_image(
                np.flipud(self.model.map_model.map), auto_level=True
            )
            self.update_dimension_cb()
            if self.widget.map_image_frame.contour_btn.isChecked():
                self._update_contours()

    def update_dimension_cb(self):
        dim_cb = self.widget.map_plot_control_widget.map_dimension_cb
        dim_cb.blockSignals(True)
        dim_cb.clear()
        possible_dimensions_str = [
            f"{x}x{y}" for x, y in self.model.map_model.possible_dimensions
        ]
        dim_cb.addItems(possible_dimensions_str)
        current_dimension_index = self.model.map_model.possible_dimensions.index(
            self.model.map_model.dimension
        )
        dim_cb.setCurrentIndex(current_dimension_index)
        dim_cb.blockSignals(False)

    def update_image(self):
        if self.model.img_model.img_data is None:
            self.widget.img_plot_widget.plot_image(np.array([[], []]))
        else:
            auto_level = self.widget.img_autoscale_btn.isChecked()
            self._setting_levels = True
            self.widget.img_plot_widget.plot_image(
                self.model.img_model.img_data, auto_level=auto_level
            )
            self._setting_levels = False
            self.plot_mask()

    def plot_mask(self):
        if self.model.current_configuration.use_mask:
            self.widget.img_plot_widget.activate_mask()
            self.widget.img_plot_widget.plot_mask(self.model.mask_model.get_mask())
        else:
            self.widget.img_plot_widget.deactivate_mask()

    def update_pattern(self):
        self.widget.pattern_plot_widget.plot_data(
            self.model.pattern.x, self.model.pattern.y, self.model.pattern.name
        )
        self.update_pattern_green_line(self.model.clicked_tth)

        cur_unit = self.model.integration_unit
        pattern_plot = self.widget.pattern_plot_widget.pattern_plot

        if cur_unit == "2th_deg":
            pattern_plot.setLabel("bottom", "2θ", "°")
            pattern_plot.invertX(False)
        elif cur_unit == "q_A^-1":
            pattern_plot.setLabel("bottom", "Q", "Å⁻¹")
            pattern_plot.invertX(False)
        elif cur_unit == "d_A":
            pattern_plot.setLabel("bottom", "d", "Å")
            pattern_plot.invertX(True)

    def update_pattern_green_line(self, pos):
        if self.model.integration_unit == "2th_deg":
            self.widget.pattern_plot_widget.set_pos_line(pos)
        else:
            wavelength = self.model.calibration_model.wavelength
            new_pos = convert_units(
                pos, wavelength, "2th_deg", self.model.integration_unit
            )
            self.widget.pattern_plot_widget.set_pos_line(new_pos)

    def update_image_green_line(self, pos):
        if not self.model.current_configuration.is_calibrated:
            return

        self.widget.img_plot_widget.set_circle_line(
            self.model.calibration_model.tth_array, np.deg2rad(pos)
        )

    def file_list_row_changed(self, row):
        self.model.map_model.select_point_by_index(row)
        self._set_stored_pattern(row)
        row, col = self.model.map_model.get_point_coordinates(row)
        map_shape = self.model.map_model.map.shape
        self.widget.map_plot_widget.set_mouse_click_position(
            col + 0.5, map_shape[0] - row - 0.5  # 0.5 are there to shift to center
        )

    def _get_mouse_row_col(self, x, y):
        x, y = np.floor(x), np.floor(y)
        row = self.widget.map_plot_widget.img_data.shape[0] - int(y) - 1
        col = int(x)
        return row, col

    def _row_col_in_map(self, row, col):
        map_shape = self.widget.map_plot_widget.img_data.shape
        if row < 0 or col < 0 or row >= map_shape[0] or col >= map_shape[1]:
            return False
        return True

    def map_point_selected(self, clicked_x, clicked_y):
        # skip when now map is loaded
        if self.model.map_model.map is None:
            return

        row, col = self._get_mouse_row_col(clicked_x, clicked_y)

        # skip when the mouse is outside of the map
        if not self._row_col_in_map(row, col):
            return

        self.model.map_model.select_point(row, col)
        ind = self.model.map_model.get_point_index(row, col)
        self._set_stored_pattern(ind)

        self.widget.control_widget.file_list.blockSignals(True)
        self.widget.control_widget.file_list.setCurrentRow(ind)
        self.widget.control_widget.file_list.blockSignals(False)

    def map_dimension_cb_changed(self, _):
        dimension_str = (
            self.widget.map_plot_control_widget.map_dimension_cb.currentText()
        )
        dimension = tuple([int(x) for x in dimension_str.split("x")])
        self.model.map_model.set_dimension(dimension)

    def map_plot_mouse_moved(self, x, y):
        # shows the information for a point inside of the map
        # since pyqtgraph gives the coordinates in the image coordinate system
        # we need to flip the y axis

        # skip when no image is loaded
        if self.widget.map_plot_widget.img_data is None:
            return

        row, col = self._get_mouse_row_col(x, y)

        # if the mouse is outside of the image, we don't want to show any information
        if not self._row_col_in_map(row, col):
            self.widget.map_plot_control_widget.mouse_x_label.setText(f"X: ")
            self.widget.map_plot_control_widget.mouse_y_label.setText(f"Y: ")
            self.widget.map_plot_control_widget.mouse_int_label.setText(f"I: ")
            self.widget.map_plot_control_widget.filename_label.setText(f"")
            return

        self.widget.map_plot_control_widget.mouse_x_label.setText(f"X: {col:.0f}")
        self.widget.map_plot_control_widget.mouse_y_label.setText(f"Y: {row:.0f}")
        self.widget.map_plot_control_widget.mouse_int_label.setText(
            f"I: {self.model.map_model.map[row, col]:.0f}"
        )

        point_info = self.model.map_model.get_point_info(row, col)
        if point_info is None:
            self.widget.map_plot_control_widget.filename_label.setText(f"")
            return

        if point_info.frame_index == 0:
            self.widget.map_plot_control_widget.filename_label.setText(
                f"{point_info.filename}"
            )
        else:
            self.widget.map_plot_control_widget.filename_label.setText(
                f"{point_info.filename} - Frame: {point_info.frame_index}"
            )

    def img_plot_left_clicked(self, x, y):
        if not self.model.current_configuration.is_calibrated:
            return

        calibration_model = self.model.calibration_model
        img_shape = self.model.img_model.img_data.shape

        x, y = np.array([y]), np.array([x])
        if 0.5 < x < img_shape[0] - 0.5 and 0.5 < y < img_shape[1] - 0.5:
            tth = calibration_model.get_two_theta_img(x, y)
            azi = calibration_model.get_azi_img(x, y)

            self.model.clicked_tth_changed.emit(np.rad2deg(tth))
            self.model.clicked_azi_changed.emit(np.rad2deg(azi))

    def img_plot_mouse_moved(self, x, y):
        image_x = self.widget.map_plot_control_widget.mouse_x_label
        image_y = self.widget.map_plot_control_widget.mouse_y_label
        image_int = self.widget.map_plot_control_widget.mouse_int_label

        if self.model.img_model.img_data is None:
            image_x.setText(f"X: ")
            image_y.setText(f"Y: ")
            image_int.setText(f"I: ")
            return

        img_shape = self.model.img_model.img_data.shape
        if 0 <= x < img_shape[1] and 0 <= y < img_shape[0]:
            image_x.setText(f"X: {x:.0f}")
            image_y.setText(f"Y: {y:.0f}")
            image_int.setText(f"I: {self.model.img_model.img_data[int(y), int(x)]:.0f}")
        else:
            image_x.setText(f"X: ")
            image_y.setText(f"Y: ")
            image_int.setText(f"I: ")

        if not self.model.current_configuration.is_calibrated:
            return

        x, y = y, x  # swap x and y for the calibration model
        img_tth = self.model.calibration_model.get_two_theta_img(x, y)
        img_tth = np.rad2deg(img_tth)
        img_azi = self.model.calibration_model.get_azi_img(x, y)
        img_azi = np.rad2deg(img_azi)

        tth_str, d_str, q_str, _ = self.get_position_strings(img_tth, "2th_deg")
        pos_widget = self.widget.pattern_footer_widget.mouse_unit_widget.cur_unit_widget
        pos_widget.tth_lbl.setText(tth_str)
        pos_widget.d_lbl.setText(d_str)
        pos_widget.q_lbl.setText(q_str)
        pos_widget.azi_lbl.setText(f"X: {img_azi:.3f}")

    def get_position_strings(
        self, x: float, current_unit: Optional[str] = None
    ) -> tuple[str, str, str, str]:
        if current_unit is None:
            current_unit = self.model.integration_unit
        if self.model.calibration_model.is_calibrated:
            wavelength = self.model.calibration_model.wavelength
            if current_unit == "2th_deg":
                tth = x
                q_value = convert_units(tth, wavelength, "2th_deg", "q_A^-1")
                d_value = convert_units(tth, wavelength, "2th_deg", "d_A")
            elif current_unit == "q_A^-1":
                q_value = x
                tth = convert_units(q_value, wavelength, "q_A^-1", "2th_deg")
                d_value = convert_units(q_value, wavelength, "q_A^-1", "d_A")
            elif current_unit == "d_A":
                d_value = x
                q_value = convert_units(d_value, wavelength, "d_A", "q_A^-1")
                tth = convert_units(d_value, wavelength, "d_A", "2th_deg")
            else:
                tth = 0
                d_value = 0
                q_value = 0

            tth_str = "2θ:%9.3f" % tth
            d_str = "d:%9.3f" % d_value
            q_str = "Q:%9.3f" % q_value
        else:
            tth_str = "2θ: -"
            d_str = "d: -"
            q_str = "Q: -"
            if current_unit == "2th_deg":
                tth_str = "2θ:%9.3f" % x
            elif current_unit == "q_A^-1":
                q_str = "Q:%9.3f" % x
            elif current_unit == "d_A":
                d_str = "d:%9.3f" % x
        azi_str = "X: -"
        return tth_str, d_str, q_str, azi_str

    def pattern_plot_mouse_moved(self, x, _):
        tth_str, d_str, q_str, azi_str = self.get_position_strings(x)
        pos_widget = self.widget.pattern_footer_widget.mouse_unit_widget.cur_unit_widget
        pos_widget.tth_lbl.setText(tth_str)
        pos_widget.d_lbl.setText(d_str)
        pos_widget.q_lbl.setText(q_str)
        pos_widget.azi_lbl.setText(azi_str)

    def update_clicked_pos_label(self, pos):
        tth_str, d_str, q_str, azi_str = self.get_position_strings(pos)
        pos_widget = (
            self.widget.pattern_footer_widget.mouse_unit_widget.clicked_unit_widget
        )
        pos_widget.tth_lbl.setText(tth_str)
        pos_widget.d_lbl.setText(d_str)
        pos_widget.q_lbl.setText(q_str)
        pos_widget.azi_lbl.setText(azi_str)

    def update_clicked_azi_label(self, azi):
        pos_widget = (
            self.widget.pattern_footer_widget.mouse_unit_widget.clicked_unit_widget
        )
        pos_widget.azi_lbl.setText(f"X: {azi:.3f}")

    def pattern_clicked(self, x, _):
        self.widget.pattern_plot_widget.map_interactive_roi.setCenter(x)

    def pattern_roi_changed(self, interactive_roi):
        region = interactive_roi.getRegion()
        self.model.map_model.set_window(region)

    def _y_scale_log_clicked(self):
        if self.widget.pattern_footer_widget.log_btn.isChecked():
            self.widget.pattern_footer_widget.sqrt_btn.setChecked(False)
            self.widget.pattern_plot_widget.set_y_scale("log")
        else:
            self.widget.pattern_plot_widget.set_y_scale("linear")

    def _y_scale_sqrt_clicked(self):
        if self.widget.pattern_footer_widget.sqrt_btn.isChecked():
            self.widget.pattern_footer_widget.log_btn.setChecked(False)
            self.widget.pattern_plot_widget.set_y_scale("sqrt")
        else:
            self.widget.pattern_plot_widget.set_y_scale("linear")

    def configuration_selected(self):
        self.update_file_list()
        self.update_map()
        self.update_image()
        self.update_pattern()



#!/usr/bin/env python
import sys
import os

from dropbox import Dropbox

if len(sys.argv) != 3:
    print('Please provide file to upload and destination folder!')
    exit(-1)

if os.getenv('DROPBOX_TOKEN') == None:
    print('Please set the Dropbox api Access token as environment variable DROPBOX_TOKEN')
    exit(-1)

db = Dropbox(os.getenv('DROPBOX_TOKEN'))
db.files_upload(open(sys.argv[1], 'rb').read(), sys.argv[2])
print('{} successfully uploaded to dropbox folder {}'.format(sys.argv[1], sys.argv[2]))



# SPDX-License-Identifier: MIT

import os
import numpy as np
import pytest

from ..utility import MockMouseEvent

unittest_data_path = os.path.join(os.path.dirname(__file__), "../data")
jcpds_path = os.path.join(unittest_data_path, "jcpds")

__all__ = ["load_proc_data", "jcpds_path", "unittest_data_path"]


@pytest.fixture
def load_proc_data(dioptas_model):
    filename = os.path.join(
        unittest_data_path, "lambda", "testasapo1_1009_00002_proc.nxs"
    )
    dioptas_model.batch_model.load_proc_data(filename)
    raw_files = dioptas_model.batch_model.files
    raw_files = [
        os.path.join(os.path.dirname(filename), os.path.basename(f)) for f in raw_files
    ]
    dioptas_model.batch_model.set_image_files(raw_files)
    return raw_files


def test_initial_state(batch_model, load_proc_data, qtbot):
    assert batch_model.n_img_all == 50
    assert batch_model.n_img == 50
    assert batch_model.binning.shape == (4038,)
    assert batch_model.data.shape == (50, 4038)


def test_is_proc(batch_controller, load_proc_data):
    filename = os.path.join(
        unittest_data_path, "lambda", "testasapo1_1009_00002_m1_part00000.nxs"
    )
    assert not batch_controller.is_proc(filename)

    filename = os.path.join(
        unittest_data_path, "lambda", "testasapo1_1009_00002_proc.nxs"
    )
    assert batch_controller.is_proc(filename)


def test_change_3d_view(batch_controller, batch_widget):
    batch_widget.activate_surface_view()

    batch_controller.set_3d_view_f()
    pg_layout = batch_widget.surface_widget.surface_view.pg_layout
    assert pg_layout.opts["azimuth"] == 0
    assert pg_layout.opts["elevation"] == 0

    batch_controller.set_3d_view_t()
    pg_layout = batch_widget.surface_widget.surface_view.pg_layout
    assert pg_layout.opts["azimuth"] == 0
    assert pg_layout.opts["elevation"] == 90

    batch_controller.set_3d_view_s()
    pg_layout = batch_widget.surface_widget.surface_view.pg_layout
    assert pg_layout.opts["azimuth"] == 90
    assert pg_layout.opts["elevation"] == 0

    batch_controller.set_3d_view_i()
    pg_layout = batch_widget.surface_widget.surface_view.pg_layout
    assert pg_layout.opts["azimuth"] == 45
    assert pg_layout.opts["elevation"] == 30


def test_wheel_event_3d(batch_controller, batch_widget, load_proc_data):
    batch_widget.activate_surface_view()
    batch_controller.change_view()

    pg_layout = batch_widget.surface_widget.surface_view.pg_layout
    show_scale = batch_widget.surface_widget.surface_view.show_scale
    show_range = batch_widget.surface_widget.surface_view.show_range
    surf_view = batch_widget.surface_widget.surface_view

    assert pg_layout.opts["distance"] == 3
    assert list(show_scale) == [2, 2, 1]
    assert list(show_range) == [0, 1]
    assert surf_view.g_translate == 0
    assert surf_view.marker == 0

    batch_controller.wheel_event_3d(MockMouseEvent())
    assert pg_layout.opts["distance"] < 10

    batch_controller.key_pressed_3d(MockMouseEvent(key=76))
    batch_controller.wheel_event_3d(MockMouseEvent())
    assert show_range[0] > 0

    batch_controller.key_pressed_3d(MockMouseEvent(key=88))
    batch_controller.wheel_event_3d(MockMouseEvent())
    assert show_scale[0] > 2.0

    batch_controller.key_pressed_3d(MockMouseEvent(key=89))
    batch_controller.wheel_event_3d(MockMouseEvent())
    assert show_scale[1] > 2.0

    batch_controller.key_pressed_3d(MockMouseEvent(key=90))
    batch_controller.wheel_event_3d(MockMouseEvent())
    assert show_scale[2] > 1.0

    batch_controller.key_pressed_3d(MockMouseEvent(key=71))
    batch_controller.wheel_event_3d(MockMouseEvent())
    assert surf_view.g_translate == 2

    batch_controller.key_pressed_3d(MockMouseEvent(key=77))
    batch_controller.wheel_event_3d(MockMouseEvent())
    assert surf_view.marker > 0


def test_pattern_left_click(
    batch_controller, pattern_controller, batch_widget, load_proc_data
):
    batch_widget.activate_stack_plot()
    batch_controller.change_view()
    assert batch_widget.stack_plot_widget.img_view.vertical_line.getXPos() == 0
    pattern_controller.pattern_left_click(15, None)
    first_pos = batch_widget.stack_plot_widget.img_view.vertical_line.getXPos()
    pattern_controller.pattern_left_click(16, None)
    second_pos = batch_widget.stack_plot_widget.img_view.vertical_line.getXPos()
    assert first_pos != second_pos


def test_subtract_background(
    batch_controller, batch_widget, batch_model, load_proc_data
):
    batch_widget.activate_stack_plot()
    batch_model.data = np.ones((100, 1000))
    batch_model.bkg = np.ones((100, 1000))
    batch_widget.options_widget.background_btn.setChecked(True)
    batch_controller.subtract_background()
    assert batch_widget.options_widget.background_btn.isChecked()
    assert np.all(
        batch_widget.stack_plot_widget.img_view.img_data
        == batch_controller.min_val["lin"]
    )


def test_extract_background(batch_controller, batch_model, load_proc_data):
    batch_model.data[...] = 1.0
    assert batch_model.bkg is None
    batch_controller.extract_background()
    assert np.allclose(batch_model.bkg, 1.0)


def test_change_scale(batch_controller, batch_widget, batch_model, load_proc_data):
    batch_widget.activate_stack_plot()
    batch_model.data[:, :] = 100

    batch_controller.change_scale_log(MockMouseEvent())
    assert batch_controller.scale == np.log10
    assert np.allclose(batch_widget.stack_plot_widget.img_view.img_data, 2.0)

    batch_controller.change_scale_sqrt(MockMouseEvent())
    assert batch_controller.scale == np.sqrt
    assert np.allclose(batch_widget.stack_plot_widget.img_view.img_data, 10.0)

    batch_controller.change_scale_lin(MockMouseEvent())
    assert batch_controller.scale == np.array
    assert np.all(batch_widget.stack_plot_widget.img_view.img_data == 100.0)


def test_process_step(batch_controller, batch_widget, batch_model, load_proc_data):
    batch_widget.position_widget.step_series_widget.stop_txt.setValue(
        batch_model.n_img - 1
    )
    # Test here only 3D plot, because waterfall is tested on functional tests
    batch_widget.activate_surface_view()
    batch_controller.plot_batch()

    assert batch_widget.surface_widget.surface_view.data.shape[0] == 50
    batch_widget.position_widget.step_series_widget.step_txt.setValue(2)
    batch_controller.process_step()
    assert batch_widget.surface_widget.surface_view.data.shape[0] == 25


def test_switch_frame(batch_controller, batch_widget, batch_model, load_proc_data):
    batch_widget.position_widget.step_series_widget.stop_txt.setValue(
        batch_model.n_img - 1
    )
    batch_widget.activate_surface_view()
    batch_controller.switch_frame(49)
    assert batch_widget.surface_widget.surface_view.g_translate == 49
    assert (
        batch_widget.position_widget.mouse_pos_widget.clicked_pos_widget.x_pos_lbl.text()
        == "Img: 49"
    )

    filename = os.path.join(
        unittest_data_path, "lambda", "testasapo1_1009_00002_m1_part00004.nxs"
    )
    assert batch_widget.windowTitle() == f"Batch widget. {filename} - 9"



# SPDX-License-Identifier: MIT

from ..utility import QtTest
import os
import gc
import numpy as np

from qtpy import QtWidgets
from mock import MagicMock
import mock

from ..utility import click_button, unittest_data_path

from ...controller.integration.CorrectionController import CorrectionController
from ...model.DioptasModel import DioptasModel
from ...widgets.integration import IntegrationWidget


class CorrectionControllerTest(QtTest):
    def setUp(self):
        self.widget = IntegrationWidget()
        self.correction_widget = self.widget.integration_control_widget.corrections_control_widget
        self.model = DioptasModel()

        self.correction_controller = CorrectionController(self.widget, self.model)

        self.original_filename = os.path.join(unittest_data_path, 'TransferCorrection', 'original.tif')
        self.response_filename = os.path.join(unittest_data_path, 'TransferCorrection', 'response.tif')

    def tearDown(self):
        del self.correction_controller
        del self.widget
        del self.model
        gc.collect()

    def load_original_img(self):
        QtWidgets.QFileDialog.getOpenFileName = MagicMock(return_value=self.original_filename)
        click_button(self.widget.transfer_load_original_btn)

    def load_response_img(self):
        QtWidgets.QFileDialog.getOpenFileName = MagicMock(return_value=self.response_filename)
        click_button(self.widget.transfer_load_response_btn)

    def test_filenames_are_displayed_in_widget(self):
        self.correction_widget.transfer_gb.setChecked(True)

        self.model.img_model.load(self.response_filename)

        self.load_original_img()
        self.assertEqual(self.correction_widget.transfer_original_filename_lbl.text(),
                         os.path.basename(self.original_filename))
        self.load_response_img()
        self.assertEqual(self.correction_widget.transfer_response_filename_lbl.text(),
                         os.path.basename(self.response_filename))

    def test_correction_loaded(self):
        self.correction_widget.transfer_gb.setChecked(True)

        self.model.img_model.load(self.response_filename)

        self.load_original_img()
        self.assertFalse(self.model.img_model.has_corrections())
        self.load_response_img()
        self.assertTrue(self.model.img_model.has_corrections())

    def test_disable_transfer_correction(self):
        self.correction_widget.transfer_gb.setChecked(True)
        self.model.img_model.load(self.response_filename)
        self.load_original_img()
        self.load_response_img()
        self.assertTrue(self.model.img_model.has_corrections())
        self.correction_widget.transfer_gb.setChecked(False)
        self.assertFalse(self.model.img_model.has_corrections())

    def test_load_img_with_different_shape(self):
        QtWidgets.QMessageBox.critical = MagicMock()
        self.correction_widget.transfer_gb.setChecked(True)
        self.model.img_model.load(self.response_filename)
        self.load_original_img()
        self.load_response_img()
        self.assertTrue(self.model.img_model.has_corrections())

        self.model.img_model.load(os.path.join(unittest_data_path, 'image_001.tif'))
        self.assertFalse(self.model.img_model.has_corrections())

        self.assertIsNone(self.model.img_model.transfer_correction.response_filename)
        self.assertFalse(self.widget.transfer_load_original_btn.isVisible())

        self.assertEqual(self.widget.transfer_original_filename_lbl.text(), 'None')
        self.assertEqual(self.widget.transfer_response_filename_lbl.text(), 'None')

    def test_load_img_with_different_shape_and_calibration(self):
        QtWidgets.QMessageBox.critical = MagicMock()
        self.correction_widget.transfer_gb.setChecked(True)
        self.model.img_model.load(self.response_filename)
        self.model.calibration_model.load(os.path.join(unittest_data_path, 'CeO2_Pilatus1M_2.poni'))

        self.load_original_img()
        self.load_response_img()
        self.assertTrue(self.model.img_model.has_corrections())

        self.model.img_model.load(os.path.join(unittest_data_path, 'image_001.tif'))
        self.assertFalse(self.model.img_model.has_corrections())

        self.assertIsNone(self.model.img_model.transfer_correction.response_filename)
        self.assertFalse(self.widget.transfer_load_original_btn.isVisible())

        self.assertEqual(self.widget.transfer_original_filename_lbl.text(), 'None')
        self.assertEqual(self.widget.transfer_response_filename_lbl.text(), 'None')

    def test_image_enable_and_disable(self):
        self.correction_widget.transfer_gb.setChecked(True)
        self.model.img_model.load(self.response_filename)
        before_data = self.model.img_data.copy()
        self.load_original_img()
        self.load_response_img()

        self.assertNotAlmostEqual(np.sum(before_data - self.model.img_data), 0)

        self.correction_widget.transfer_gb.setChecked(False)
        self.assertAlmostEqual(np.sum(before_data - self.model.img_data), 0)

    def test_image_enable_and_disable_with_calibration(self):
        self.correction_widget.transfer_gb.setChecked(True)
        self.model.img_model.load(self.response_filename)
        self.model.calibration_model.load(os.path.join(unittest_data_path, 'CeO2_Pilatus1M_2.poni'))
        before_data = self.model.img_data.copy()
        self.load_original_img()
        self.load_response_img()

        self.assertNotAlmostEqual(np.sum(before_data - self.model.img_data), 0)

        self.correction_widget.transfer_gb.setChecked(False)
        self.assertAlmostEqual(np.sum(before_data - self.model.img_data), 0)

        self.correction_widget.transfer_gb.setChecked(True)
        self.assertNotAlmostEqual(np.sum(before_data - self.model.img_data), 0)

    def test_show_correction_in_img_widget_and_back(self):
        self.correction_widget.transfer_gb.setChecked(True)
        self.model.img_model.load(self.response_filename)
        self.load_original_img()
        self.load_response_img()

        click_button(self.correction_widget.transfer_plot_btn)
        transfer_data = self.model.img_model.transfer_correction.transfer_data
        img_data = self.widget.img_widget.img_data
        self.assertAlmostEqual(np.sum(transfer_data - img_data), 0)
        self.assertTrue(self.correction_widget.transfer_plot_btn.isChecked())
        self.assertEqual(self.correction_widget.transfer_plot_btn.text(), 'Back')

        click_button(self.correction_widget.transfer_plot_btn)

        self.assertFalse(self.correction_widget.transfer_plot_btn.isChecked())
        self.assertEqual(self.correction_widget.transfer_plot_btn.text(), 'Plot')

    def test_transfer_correction_is_applied_correctly(self):
        self.model.calibration_model.load(os.path.join(unittest_data_path, 'TransferCorrection', 'transfer.poni'))
        self.model.img_model.load(self.original_filename)
        y_original = self.model.pattern.y
        self.model.img_model.load(self.response_filename)

        self.correction_widget.transfer_gb.setChecked(True)
        self.load_original_img()
        self.load_response_img()
        y_response_with_transfer = self.model.pattern.y

        self.assertAlmostEqual(np.sum(y_original-y_response_with_transfer), 0)

    def test_changing_transfer_function_several_times(self):
        self.model.img_model.load(self.original_filename)
        self.correction_widget.transfer_gb.setChecked(True)
        self.load_original_img()
        self.load_response_img()

        img_1 = self.model.img_data.copy()

        QtWidgets.QFileDialog.getOpenFileName = MagicMock(return_value=self.original_filename)
        click_button(self.widget.transfer_load_response_btn)

        img_2 = self.model.img_data.copy()

        self.assertFalse(np.array_equal(img_1, img_2))








# SPDX-License-Identifier: MIT

import os
import numpy as np
from qtpy import QtWidgets
from mock import MagicMock
from pytest import approx

from dioptas.controller.MainController import MainController

from ..utility import click_button

unittest_path = os.path.dirname(__file__)
data_path = os.path.join(unittest_path, os.pardir, "data")
map_img_path = os.path.join(data_path, "map")
map_pattern_path = os.path.join(data_path, "map", "xy")
map_img_file_names = [
    f for f in os.listdir(map_img_path) if os.path.isfile(os.path.join(map_img_path, f))
]
map_img_file_paths = [
    os.path.join(map_img_path, filename) for filename in map_img_file_names
]


def mock_open_filenames(filepaths):
    QtWidgets.QFileDialog.getOpenFileNames = MagicMock(return_value=filepaths)


def prepare_map_gui(main_controller: MainController):
    main_controller.model.current_configuration.img_model.load(map_img_file_paths[0])
    main_controller.model.current_configuration.calibration_model.load(
        os.path.join(data_path, "CeO2_Pilatus1M.poni")
    )
    map_widget = main_controller.widget.map_widget
    click_button(main_controller.widget.map_mode_btn)
    mock_open_filenames(map_img_file_paths)
    click_button(map_widget.control_widget.load_btn)


def test_map(main_controller: MainController):
    # Herbert has recently collected a large map of his sample and wants to visualize
    # and explore it in Dioptas.

    # He opens Dioptas loads his calibration file and is curious to see that there
    # is a mode on the left side which is called "Map".
    main_controller.model.current_configuration.img_model.load(map_img_file_paths[0])
    main_controller.model.current_configuration.calibration_model.load(
        os.path.join(data_path, "CeO2_Pilatus1M.poni")
    )

    # He clicks on the map mode and sees a beautiful gui with 2 image views and one pattern view.

    click_button(main_controller.widget.map_mode_btn)
    assert main_controller.widget.map_widget.isVisible()

    map_widget = main_controller.widget.map_widget

    assert map_widget.map_pg_layout.isVisible()
    assert map_widget.img_pg_layout.isVisible()
    assert map_widget.pattern_pg_layout.isVisible()

    # He realizes that there is also a control widget on the rightside which allows him to load
    # images. He clicks on the load button and load a list of files.

    assert map_widget.control_widget.isVisible()
    mock_open_filenames(map_img_file_paths)

    click_button(map_widget.control_widget.load_btn)
    # The loading of the files generates a list of files in the GUI. Further, it automatically
    # generates a map of something on the left side of the GUI.

    assert map_widget.control_widget.file_list.count() == len(map_img_file_names)
    file_list_strings = [
        map_widget.control_widget.file_list.item(i).text()
        for i in range(map_widget.control_widget.file_list.count())
    ]
    assert map_img_file_names == file_list_strings
    assert map_widget.map_plot_widget.img_data is not None

    # He notices that the first image is loaded in the corresponding image view
    # and the corresponding pattern is shown in the pattern view.

    assert map_widget.img_plot_widget.img_data is not None
    assert map_widget.pattern_plot_widget.plot_item.getData() is not None

    # He clicks on the file list in the control widget, and sees, that the image and pattern
    # is updated in the corresponding views.

    map_widget.control_widget.file_list.setCurrentRow(1)
    loaded_filename = main_controller.model.current_configuration.img_model.filename
    assert loaded_filename == map_img_file_paths[1]

    # Moving his mouse over the map on the right, he sees that the x, y indices of the map
    # are shown on the bottom of the map.

    map_widget.map_plot_widget.mouse_moved.emit(2, 1)
    assert map_widget.map_plot_control_widget.mouse_x_label.text() == "X: 2"
    assert map_widget.map_plot_control_widget.mouse_y_label.text() == "Y: 1"

    # He clicks on individual points of the map and sees that the corresponding image and pattern
    # is shown in the right side of the GUI.
    map_widget.map_plot_widget.mouse_left_clicked.emit(2, 2)
    loaded_filename = main_controller.model.current_configuration.img_model.filename
    assert loaded_filename == map_img_file_paths[2]
    assert map_widget.control_widget.file_list.currentRow() == 2

    # Suddenly he realizes that there is an interactive ROI in the pattern view.
    assert (
        map_widget.pattern_plot_widget.map_interactive_roi
        in map_widget.pattern_plot_widget.pattern_plot.items
    )

    # click anywhere in the pattern view moves the roi and updates the map
    pattern = main_controller.model.current_configuration.pattern_model.pattern
    center_x = np.mean(pattern.x)
    center_y = np.mean(pattern.y)
    map_widget.pattern_plot_widget.mouse_left_clicked.emit(center_x, center_y)

    assert map_widget.pattern_plot_widget.map_interactive_roi.center == approx(center_x)


def test_map_with_different_dimension(main_controller: MainController):
    # Herbert has collected another large map of his sample and wants to visualize
    # and explore it in Dioptas. This time the map is not rectangular but has a
    # different dimension (4x2)

    # He opens Dioptas loads his calibration file and is curious to see that there
    # is a mode on the left side which is called "Map".
    main_controller.model.current_configuration.img_model.load(map_img_file_paths[0])
    main_controller.model.current_configuration.calibration_model.load(
        os.path.join(data_path, "CeO2_Pilatus1M.poni")
    )
    map_widget = main_controller.widget.map_widget

    # He clicks on the map mode and loads his images.

    click_button(main_controller.widget.map_mode_btn)
    mock_open_filenames(map_img_file_paths[0:8])

    click_button(map_widget.control_widget.load_btn)

    # He realizes that the resulting map has a 4x2 dimension.

    assert map_widget.map_plot_widget.img_data.shape == (2, 4)

    # however, it is obvious that he actually needs a 4x2 map. He sees, that there are
    # multiple possible dimensions available below the image widget. He clicks on the
    # 2x4 dimension and sees that the map is updated.

    assert map_widget.map_plot_control_widget.map_dimension_cb.currentText() == "2x4"
    dim_cb = map_widget.map_plot_control_widget.map_dimension_cb
    dim_cb_str_list = [dim_cb.itemText(i) for i in range(dim_cb.count())]
    dim_model_str_list = [
        f"{x}x{y}"
        for x, y in main_controller.model.current_configuration.map_model.possible_dimensions
    ]
    assert dim_cb.count() == 4
    assert dim_cb_str_list == dim_model_str_list

    index_4x2 = dim_cb_str_list.index("4x2")
    dim_cb.setCurrentIndex(index_4x2)
    assert map_widget.map_plot_widget.img_data.shape == (4, 2)

    # He is satisfied and can go on exploring his map of the sample.
